"""Create an association based on skycells"""

import argparse
import json
import logging
import sys

import numpy as np
import roman_datamodels as rdm

import romancal.patch_match.patch_match as pm
from romancal.associations import asn_from_list
from romancal.lib.basic_utils import parse_visitID as parse_visitID

__all__ = ["skycell_asn"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.setLevel("INFO")


def skycell_asn(self):
    """Create the associaton from the list"""
    all_patches = []
    file_list = []
    for file_name in self.parsed.filelist:
        cal_file = rdm.open(file_name)
        filter_id = cal_file.meta.instrument.optical_element.lower()
        file_patch_list = pm.find_patch_matches(cal_file.meta.wcs)
        logger.info(f"Patch List:{file_name}, {file_patch_list[0]}")
        file_list.append([file_name, file_patch_list[0]])
        all_patches.append(file_patch_list[0])

    unique_patches = np.unique(np.concatenate(all_patches))
    for item in unique_patches:
        member_list = []
        patch_name = pm.PATCH_TABLE[item]["name"]
        for a in file_list:
            if np.isin(item, a[1]):
                member_list.append(a[0])

        # grab all the wcs parameters needed for generate_tan_wcs
        projcell_info = dict(
            [
                ("pixel_scale", float(pm.PATCH_TABLE[item]["pixel_scale"])),
                ("ra_cent", float(pm.PATCH_TABLE[item]["ra_projection_center"])),
                ("dec_cent", float(pm.PATCH_TABLE[item]["dec_projection_center"])),
                ("shiftx", float(pm.PATCH_TABLE[item]["x0_projection"])),
                ("shifty", float(pm.PATCH_TABLE[item]["y0_projection"])),
                ("nx", int(pm.PATCH_TABLE[item]["nx"])),
                ("ny", int(pm.PATCH_TABLE[item]["ny"])),
                ("orient", float(pm.PATCH_TABLE[item]["orientat"])),
                (
                    "orientat_projection_center",
                    float(pm.PATCH_TABLE[item]["orientat_projection_center"]),
                ),
            ]
        )
        parsed_visit_id = parse_visitID(member_list[0][1:20])
        program_id = parsed_visit_id["Program"]
        root_asn_name = self.parsed.output_file_root
        product_type = self.parsed.product_type
        product_release = self.parsed.release_product
        suffix = "coadd"
        sep = "_"

        product_name_mapping = {
            "visit": "v"
            + parsed_visit_id["Execution"]
            + parsed_visit_id["Pass"]
            + parsed_visit_id["Segment"]
            + parsed_visit_id["Observation"],
            "daily": "d"
            + parsed_visit_id["Execution"]
            + parsed_visit_id["Pass"]
            + parsed_visit_id["Segment"],
            "pass": "p" + parsed_visit_id["Execution"] + parsed_visit_id["Pass"],
            "full": "full",
            "user": "user",
        }

        pr_name = product_name_mapping.get(product_type, "unknown")

        asn_file_name = (
            root_asn_name
            + sep
            + product_release
            + sep
            + pr_name
            + sep
            + patch_name
            + sep
            + filter_id
            + sep
            + suffix
        )

        with open(asn_file_name + "_asn.json", "w") as outfile:
            prompt_product_asn = asn_from_list.asn_from_list(
                member_list, product_name=asn_file_name
            )
            prompt_product_asn["asn_type"] = "image"
            prompt_product_asn["program"] = program_id
            prompt_product_asn["target"] = patch_name
            prompt_product_asn["skycell_wcs_info"] = json.dumps(projcell_info)
            _, serialized = prompt_product_asn.dump(format="json")
            outfile.write(serialized)


class Main:
    """Command-line interface for list_to_asn

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """

    def __init__(self, args=None):
        if args is None:
            args = sys.argv[1:]
        if isinstance(args, str):
            args = args.split(" ")

        parser = argparse.ArgumentParser(
            description="Create an association from a list of files",
            usage="skycell_asn --product-type visit --release-product prompt *_cal.asdf -o r512",
        )

        parser.add_argument(
            "-o",
            "--output-file-root",
            type=str,
            required=True,
            help="Root string for file to write association to",
        )

        parser.add_argument(
            "-f",
            "--format",
            type=str,
            default="json",
            help='Format of the association files. Default: "%(default)s"',
        )

        parser.add_argument(
            "--product-type",
            type=str,
            default="visit",
            help="The product type when creating the association",
        )

        parser.add_argument(
            "--release-product",
            type=str,
            default="p",
            help="The release product when creating the association",
        )

        parser.add_argument(
            "-r",
            "--rule",
            type=str,
            default="DMS_ELPP_Base",
            help=(
                "The rule to base the association structure on."
                ' Default: "%(default)s"'
            ),
        )
        parser.add_argument(
            "-i",
            "--id",
            type=str,
            default="o999",
            help='The association candidate id to use. Default: "%(default)s"',
            dest="acid",
        )

        parser.add_argument(
            "filelist",
            type=str,
            nargs="+",
            help="File list to include in the association",
        )

        self.parsed = parser.parse_args(args=args)
        logger.info("Command-line arguments: %s", self.parsed)

        skycell_asn(self)


if __name__ == "__main__":

    Main()



"""
Utilities for fitting model PSFs to rate images.
"""

import logging

import numpy as np
import webbpsf
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.nddata import bitmask
from astropy.table import Table
from photutils.background import LocalBackground
from photutils.detection import DAOStarFinder
from photutils.psf import IterativePSFPhotometry, PSFPhotometry, SourceGrouper
from roman_datamodels.datamodels import ImageModel
from roman_datamodels.dqflags import pixel
from webbpsf import gridded_library

__all__ = [
    "create_gridded_psf_model",
    "fit_psf_to_image_model",
    "dq_to_boolean_mask",
]

# set loggers to debug level by default:
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

# Phase C central wavelengths [micron], released by Goddard (Jan 2023):
# https://roman.ipac.caltech.edu/sims/Param_db.html#wfi_filters
filter_central_wavelengths = {
    "F062": 0.620,
    "F087": 0.869,
    "F106": 1.060,
    "F129": 1.293,
    "F146": 1.464,
    "F158": 1.577,
    "F184": 1.842,
    "F213": 2.125,
}

default_finder = DAOStarFinder(
    # these defaults extracted from the
    # romancal SourceDetectionStep
    fwhm=1.0,
    threshold=0.0,
    sharplo=0.0,
    sharphi=1.0,
    roundlo=-1.0,
    roundhi=1.0,
    peakmax=None,
)


def create_gridded_psf_model(
    filt,
    detector,
    oversample=11,
    fov_pixels=9,
    sqrt_n_psfs=2,
    buffer_pixels=100,
    instrument_options=None,
):
    """
    Compute a gridded PSF model for one SCA via
    `~webbpsf.gridded_library.CreatePSFLibrary`.

    Parameters
    ----------
    filt : str
        Filter name, starting with "F". For example: `"F184"`.
    detector : str
        Computed gridded PSF model for this SCA.
        Examples include: `"SCA01"` or `"SCA18"`.
    oversample : int, optional
        Oversample factor, default is 11. See WebbPSF docs for details [1]_.
        Choosing an odd number makes the pixel convolution more accurate.
    fov_pixels : int, optional
        Field of view width [pixels]. Default is 12.
        See WebbPSF docs for details [1]_.
    sqrt_n_psfs : int, optional
        Square root of the number of PSFs to calculate, distributed uniformly
        across the detector. Default is 4.
    buffer_pixels : int, optional
        Calculate a grid of PSFs distributed uniformly across the detector
        at least ``buffer_pixels`` away from the detector edges. Default is 100.
    instrument_options : dict, optional
        Instrument configuration options passed to WebbPSF.
        For example, WebbPSF assumes Roman pointing jitter consistent with
        mission specs by default, but this can be turned off with:
        ``{'jitter': None, 'jitter_sigma': 0}``.

    Returns
    -------
    gridmodel : `photutils.psf.GriddedPSFModel`
        Gridded PSF model evaluated at several locations on one SCA.
    model_psf_centroids : list of tuples
        Pixel locations of the PSF models calculated for ``gridmodel``.

    References
    ----------
    .. [1] `WebbPSF documentation for `webbpsf.JWInstrument.calc_psf`
       <https://webbpsf.readthedocs.io/en/latest/api/webbpsf.JWInstrument.html#webbpsf.JWInstrument.calc_psf>`_

    """
    if int(sqrt_n_psfs) != sqrt_n_psfs:
        raise ValueError(f"`sqrt_n_psfs` must be an integer, got {sqrt_n_psfs}.")
    n_psfs = int(sqrt_n_psfs) ** 2

    # Choose pixel boundaries for the grid of PSFs:
    start_pix = 0
    stop_pix = 4096

    # Choose locations on detector for each PSF:
    if sqrt_n_psfs != 1:
        pixel_range = np.linspace(
            start_pix + buffer_pixels, stop_pix - buffer_pixels, int(sqrt_n_psfs)
        )
    else:
        pixel_range = [(start_pix + stop_pix) / 2]

    # generate PSFs over a grid of detector positions [pix]
    model_psf_centroids = [(int(x), int(y)) for y in pixel_range for x in pixel_range]

    wfi = webbpsf.roman.WFI()
    wfi.filter = filt

    if instrument_options is not None:
        wfi.options.update(instrument_options)

    # Initialize the PSF library
    inst = gridded_library.CreatePSFLibrary(
        instrument=wfi,
        filter_name=filt,
        detectors=detector.upper(),
        num_psfs=n_psfs,
        oversample=oversample,
        fov_pixels=fov_pixels,
        add_distortion=False,
        crop_psf=False,
        save=False,
        verbose=False,
    )

    inst.location_list = model_psf_centroids

    # Create the PSF grid:
    gridmodel = inst.create_grid()

    return gridmodel, model_psf_centroids


def fit_psf_to_image_model(
    image_model=None,
    data=None,
    error=None,
    dq=None,
    photometry_cls=PSFPhotometry,
    psf_model=None,
    grouper=None,
    fitter=None,
    localbkg_estimator=None,
    finder=None,
    x_init=None,
    y_init=None,
    progress_bar=False,
    error_lower_limit=None,
    fit_shape=(15, 15),
    exclude_out_of_bounds=True,
):
    """
    Fit PSF models to an ``ImageModel``.

    Parameters
    ----------
    image_model : `roman_datamodels.datamodels.ImageModel`
        Image datamodel. If ``image_model`` is supplied,
        ``data,error,dq`` should be `None`.
    data : `astropy.units.Quantity`
        Fit a PSF model to the rate image ``data``.
        If ``data,error,dq`` are supplied, ``image_model`` should be `None`.
    error : `astropy.units.Quantity`
        Uncertainties on fluxes in ``data``. Should be `None` if
        ``image_model`` is supplied.
    dq : `numpy.ndarray`
        Data quality bitmask for ``data``. Should be `None` if
        ``image_model`` is supplied.
    photometry_cls : {`photutils.psf.PSFPhotometry`,
            `photutils.psf.IterativePSFPhotometry`}
        Choose a photutils PSF photometry technique (default or iterative).
    psf_model : `astropy.modeling.Fittable2DModel`
        The 2D PSF model to fit to the rate image. Usually this model is an instance
        of `photutils.psf.GriddedPSFModel`.
    grouper : `photutils.psf.SourceGrouper`
        Specifies rules for attempting joint fits of multiple PSFs when
         there are nearby sources at small separations.
    fitter : `astropy.modeling.fitting.Fitter`, optional
        Modeling class which optimizes the PSF fit.
        Default is `astropy.modeling.fitting.LevMarLSQFitter(calc_uncertainties=True)`.
    localbkg_estimator : `photutils.background.LocalBackground`, optional
        Specifies inner and outer radii for computing flux background near
        a source. Default has ``inner_radius=10, outer_radius=30``.
    finder : subclass of `photutils.detection.StarFinderBase`, optional
        When ``photutils_cls`` is `photutils.psf.IterativePSFPhotometry`, the
        ``finder`` is called to determine if sources remain in the rate image
        after one PSF model is fit to the observations and removed.
        Default was extracted from the `DAOStarFinder` call in the
        Source Detection step.
    x_init : `numpy.ndarray`, optional
        Initial guesses for the ``x`` pixel coordinates of each source to fit.
    y_init : `numpy.ndarray`, optional
        Initial guesses for the ``y`` pixel coordinates of each source to fit.
    progress_bar : bool, optional
        Render a progress bar via photutils. Default is False.
    error_lower_limit : `astropy.units.Quantity`, optional
        Since some synthetic images may have bright sources with very
        small statistical uncertainties, the ``error`` can be clipped at
        ``error_lower_limit`` to prevent over-confident fits.
    fit_shape : int, or tuple of length 2, optional
        Rectangular shape around the center of a star that will
        be used to define the PSF-fitting data. See docs for
        `photutils.psf.PSFPhotometry` for details. Default is ``(16, 16)``.
    exclude_out_of_bounds : bool, optional
        If `True`, do not attempt to fit stars which have initial centroids
        that fall outside the pixel limits of the SCA. Default is False.

    Returns
    -------
    results_table : `astropy.table.QTable`
        PSF photometry results.
    photometry : instance of class ``photutils_cls``
        PSF photometry instance with configuration settings and results.

    """
    if grouper is None:
        # minimum separation before sources are fit simultaneously:
        grouper = SourceGrouper(min_separation=5)  # [pix]

    if fitter is None:
        fitter = LevMarLSQFitter(calc_uncertainties=True)

    # the iterative PSF method requires a finder:
    psf_photometry_kwargs = {}
    if photometry_cls is IterativePSFPhotometry or (x_init is None and y_init is None):
        if finder is None:
            finder = default_finder
        psf_photometry_kwargs["finder"] = finder

    if localbkg_estimator is None:
        localbkg_estimator = LocalBackground(
            inner_radius=10,  # [pix]
            outer_radius=30,  # [pix]
        )

    photometry = photometry_cls(
        grouper=grouper,
        localbkg_estimator=localbkg_estimator,
        psf_model=psf_model,
        fitter=fitter,
        fit_shape=fit_shape,
        aperture_radius=fit_shape[0],
        progress_bar=progress_bar,
        **psf_photometry_kwargs,
    )

    if x_init is not None and y_init is not None:
        guesses = Table(np.column_stack([x_init, y_init]), names=["x_init", "y_init"])
    else:
        guesses = None

    if image_model is None:
        if data is None and error is None:
            raise ValueError(
                "PSF fitting requires either an ImageModel, "
                "or arrays for the data and error."
            )

    ignore_flags = pixel.NO_LIN_CORR
    # presently the linearity correction is somewhat problematic in
    # CRDS reference files; we should replace this with ignore_flags = 0
    # at some point in the future.

    if dq is None:
        if image_model is not None and isinstance(image_model, ImageModel):
            # L2 images have a dq array
            mask = dq_to_boolean_mask(image_model, ignore_flags=ignore_flags)
        else:
            # L3 images
            mask = image_model.weight == 0
    else:
        mask = dq_to_boolean_mask(dq)

    if data is None and image_model is not None:
        data = image_model.data

    if error is None and image_model is not None:
        error = image_model.err

    if error_lower_limit is not None:
        # option to enforce a lower limit on the flux uncertainties
        error = np.clip(error, error_lower_limit, None)

    # we also mask non-finite values in the data and error arrays:
    non_finite = ~np.isfinite(data) | ~np.isfinite(error)

    if exclude_out_of_bounds and guesses is not None:
        # don't attempt to fit PSFs for objects with initial centroids
        # outside the detector boundaries:
        init_centroid_in_range = (
            (guesses["x_init"] > 0)
            & (guesses["x_init"] < data.shape[1])
            & (guesses["y_init"] > 0)
            & (guesses["y_init"] < data.shape[0])
        )
        guesses = guesses[init_centroid_in_range]

    # fit the model PSF to the data:
    results_table = photometry(
        data=data, error=error, init_params=guesses, mask=mask | non_finite
    )

    # results are stored on the PSFPhotometry instance:
    return results_table, photometry


def dq_to_boolean_mask(image_model_or_dq, ignore_flags=0, flag_map_name="ROMAN_DQ"):
    """
    Convert a DQ bitmask to a boolean mask. Useful for photutils methods.

    Parameters
    ----------
    image_model_or_dq : `roman_datamodels.datamodels.ImageModel` or `numpy.ndarray`
        ImageModel containing the DQ bitmask to convert to a boolean mask,
        or the DQ bitmask itself.
    ignore_flags : int, str, list, None (default = 0)
        See docs for `astropy.nddata.bitmask.extend_bit_flag_map`
    flag_map_name : str
        Name for the bitmask flag map in the astropy bitmask registry

    Returns
    -------
    mask : `numpy.ndarray`
        Boolean mask
    """

    if isinstance(image_model_or_dq, ImageModel):
        dq = image_model_or_dq.dq
    else:
        dq = image_model_or_dq

    # add the Roman DQ flags to the astropy bitmask registry:
    dq_flag_map = {dq.name: dq.value for dq in pixel if dq.name != "GOOD"}
    bitmask.extend_bit_flag_map(flag_map_name, **dq_flag_map)

    # convert the bitmask to a boolean mask:
    mask = bitmask.bitfield_to_boolean_mask(dq, ignore_flags=ignore_flags)
    return mask.astype(bool)



__all__ = [
    "AssociationError",
    "AssociationNotAConstraint",
    "AssociationNotValidError",
]


class AssociationError(Exception):
    """Basic errors related to Associations"""


class AssociationNotAConstraint(AssociationError):
    """No matching constraint found"""


class AssociationNotValidError(AssociationError):
    """Given data structure is not a valid association"""



"""Base classes for Roman tests"""

import os
import sys
from glob import glob as _sys_glob
from pathlib import Path

import requests

# Define location of default Artifactory API key, for Jenkins use only
ARTIFACTORY_API_KEY_FILE = "/eng/ssb2/keys/svc_rodata.key"


# Define a request timeout in seconds
TIMEOUT = 30


def _data_glob_local(*glob_parts):
    """Perform a glob on the local path

    Parameters
    ----------
    glob_parts: (path-like,[...])
        List of components that will be built into a single path

    Returns
    -------
    file_paths: [str[, ...]]
        Full file paths that match the glob criterion
    """
    full_glob = Path().joinpath(*glob_parts)
    return _sys_glob(str(full_glob))


def _data_glob_url(*url_parts, root=None):
    """
    Parameters
    ----------
    url: (str[,...])
        List of components that will be used to create a URL path

    root: str
        The root server path to the Artifactory server.
        Normally retrieved from `get_bigdata_root`.

    Returns
    -------
    url_paths: [str[, ...]]
        Full URLS that match the glob criterion
    """
    # Fix root root-ed-ness
    if root.endswith("/"):
        root = root[:-1]

    # Access
    try:
        envkey = os.environ["API_KEY_FILE"]
    except KeyError:
        envkey = ARTIFACTORY_API_KEY_FILE

    try:
        with open(envkey) as fp:
            headers = {"X-JFrog-Art-Api": fp.readline().strip()}
    except (PermissionError, FileNotFoundError):
        print(
            "Warning: Anonymous Artifactory search requests are limited "
            "to 1000 results. Use an API key and define API_KEY_FILE "
            "environment variable to get full search results.",
            file=sys.stderr,
        )
        headers = None

    search_url = "/".join([root, "api/search/pattern"])

    # Join and re-split the url so that every component is identified.
    url = "/".join([root] + [idx for idx in url_parts])
    all_parts = url.split("/")

    # Pick out "roman-pipeline", the repo name
    repo = all_parts[4]

    # Format the pattern
    pattern = repo + ":" + "/".join(all_parts[5:])

    # Make the query
    params = {"pattern": pattern}
    with requests.get(search_url, params=params, headers=headers, timeout=TIMEOUT) as r:
        url_paths = r.json()["files"]

    return url_paths



"""
 Unit tests for the Roman source detection step code
"""

from copy import deepcopy

import numpy as np
import pytest
from astropy import units as u

from romancal.lib.basic_utils import recarray_to_ndarray
from romancal.lib.tests.test_psf import add_sources, setup_inputs
from romancal.source_detection import SourceDetectionStep

n_sources = 10
image_model_shape = (100, 100)


class TestSourceDetection:
    def setup_method(self):
        self.image_model, self.webbpsf_config, self.psf_model = setup_inputs(
            shape=image_model_shape,
        )

    @pytest.mark.webbpsf
    def test_dao_vs_psf(self, seed=0):
        rng = np.random.default_rng(seed)
        image_model = deepcopy(self.image_model)

        n_models = 10
        flux_true = rng.normal(1e3, 100, n_models)
        along_diag = np.arange(100, 950, 90) / 1000 * image_model_shape[0]
        x_true = along_diag + rng.normal(scale=0.5, size=n_models)
        y_true = along_diag + rng.normal(scale=0.5, size=n_models)

        add_sources(image_model, self.psf_model, x_true, y_true, flux_true)

        source_detect = SourceDetectionStep()
        source_detect.scalar_threshold = 100
        source_detect.peakmax = None
        source_detect.fit_psf = False
        dao_result = source_detect.process(image_model)
        idx, x_dao, y_dao, amp_dao = recarray_to_ndarray(
            dao_result.meta.source_detection.tweakreg_catalog
        ).T

        # check that all injected targets are found by DAO:
        assert len(x_dao) == len(x_true)

        source_detect.fit_psf = True
        psf_result = source_detect.process(image_model)
        psf_catalog = psf_result.meta.source_detection.tweakreg_catalog

        extract_columns = ["xcentroid", "x_err", "ycentroid", "y_err", "flux_fit"]
        x_psf, x_err, y_psf, y_err, amp_psf = psf_catalog[extract_columns].itercols()

        # check that typical PSF centroids are more accurate than DAO centroids:
        assert np.median(np.abs(x_dao - x_true)) > np.median(np.abs(x_psf - x_true))

        # check that the typical/worst PSF centroid is still within some tolerance:
        pixel_scale = 0.11 * u.arcsec / u.pix
        centroid_residuals = np.abs(x_psf - x_true) * u.pix * pixel_scale
        assert np.max(centroid_residuals) < 11 * u.mas
        assert np.median(centroid_residuals) < 3 * u.mas

        # check that typical residuals are consistent with their errors:
        assert np.median(np.abs(x_psf - x_true) / x_err) < 2



import logging
import warnings
from typing import Tuple

import gwcs
import numpy as np
from astropy import wcs as fitswcs
from astropy.modeling import Model
from astropy.nddata.bitmask import bitfield_to_boolean_mask
from roman_datamodels.dqflags import pixel
from stcal.alignment.util import wcs_from_footprints

from romancal.assign_wcs.utils import wcs_bbox_from_shape

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def make_output_wcs(
    input_models,
    pscale_ratio=None,
    pscale=None,
    rotation=None,
    shape=None,
    crpix: Tuple[float, float] = None,
    crval: Tuple[float, float] = None,
):
    """
    Generate output WCS here based on footprints of all input WCS objects

    Parameters
    ----------
    input_models : list of `roman_datamodels.datamodels.DataModel`
        Each datamodel must have a `gwcs.wcs.WCS` object.

    pscale_ratio : float, optional
        Ratio of input to output pixel scale. Ignored when ``pscale`` is provided.

    pscale : float, None, optional
        Absolute pixel scale in degrees. When provided, overrides ``pscale_ratio``.

    rotation : float, None, optional
        Position angle (in degrees) of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes.

    shape : tuple of int, None, optional
        Shape of the image (data array) using ``numpy.ndarray`` convention
        (``ny`` first and ``nx`` second). This value will be assigned
        to ``pixel_shape`` and ``array_shape`` properties of the returned
        WCS object.

    crpix : tuple of float, None, optional
        Position of the reference pixel in the image array.  If ``ref_pixel`` is not
        specified, it will be set to the center of the bounding box of the
        returned WCS object.

    crval : tuple of float, None, optional
        Right ascension and declination of the reference pixel. Automatically
        computed if not provided.

    Returns
    -------
    output_wcs : object
        WCS object, with defined domain, covering entire set of input frames
    """

    wcslist = [i.meta.wcs for i in input_models]
    for w, i in zip(wcslist, input_models):
        if w.bounding_box is None:
            w.bounding_box = wcs_bbox_from_shape(i.data.shape)
    naxes = wcslist[0].output_frame.naxes

    if naxes != 2:
        raise RuntimeError(f"Output WCS needs 2 axes.{wcslist[0]} has {naxes}.")

    output_wcs = wcs_from_footprints(
        wcslist,
        None,
        dict(input_models[0].meta.wcsinfo),
        pscale_ratio=pscale_ratio,
        pscale=pscale,
        rotation=rotation,
        shape=shape,
        crpix=crpix,
        crval=crval,
    )

    # Check that the output data shape has no zero length dimensions
    if not np.prod(output_wcs.array_shape):
        raise ValueError(
            f"Invalid output frame shape: {tuple(output_wcs.array_shape)}; dimensions"
            " must have length > 0."
        )

    return output_wcs


def build_driz_weight(
    model,
    weight_type=None,
    good_bits: str = None,
):
    """
    Builds the drizzle weight map for resampling.

    Parameters
    ----------
    model : object
        The input model.
    weight_type : str, optional
        The type of weight to use. Allowed values are 'ivm' or 'exptime'.
        Defaults to None.
    good_bits : str, optional
        The good bits to use for building the mask. Defaults to None.

    Returns
    -------
    numpy.ndarray
        The drizzle weight map.

    Raises
    ------
    ValueError
        If an invalid weight type is provided.

    Examples
    --------
    .. code-block:: none

        model = get_input_model()
        weight_map = build_driz_weight(model, weight_type='ivm', good_bits=[1, 2, 3])
        print(weight_map)
    """

    dqmask = bitfield_to_boolean_mask(
        model.dq,
        good_bits,
        good_mask_value=1,
        dtype=np.uint8,
        flag_name_map=pixel,
    )

    if weight_type == "ivm":
        if (
            hasattr(model, "var_rnoise")
            and model.var_rnoise is not None
            and model.var_rnoise.shape == model.data.shape
        ):
            with np.errstate(divide="ignore", invalid="ignore"):
                inv_variance = model.var_rnoise**-1
            inv_variance[~np.isfinite(inv_variance)] = 1
        else:
            warnings.warn(
                "var_rnoise array not available. Setting drizzle weight map to 1",
                RuntimeWarning,
            )
            inv_variance = 1.0
        result = inv_variance * dqmask
    elif weight_type == "exptime":
        exptime = model.meta.exposure.exposure_time
        result = exptime * dqmask
    elif weight_type is None:
        result = np.ones(model.data.shape, dtype=model.data.dtype) * dqmask
    else:
        raise ValueError(
            f"Invalid weight type: {weight_type}."
            "Allowed weight types are 'ivm' or 'exptime'."
        )

    return result.astype(np.float32)


def build_mask(dqarr, bitvalue):
    """
    Build a bit mask from an input DQ array and a bitvalue flag.

    Parameters
    ----------
    dqarr : numpy.ndarray
        Input DQ array.
    bitvalue : str
        Bitvalue flag.

    Returns
    -------
    ndarray
        Bit mask where 1 represents good and 0 represents bad.

    Notes
    -----
    - The function interprets the bitvalue flag using the
      `astropy.nddata.bitmask.interpret_bit_flags` function.
    - If the bitvalue is None, the function returns a bit mask with all elements
      set to 1.
    - Otherwise, the function performs a bitwise AND operation between the dqarr and
      the complement of the bitvalue, and then applies a logical NOT operation to
      obtain the bit mask.
    - The resulting bit mask is returned as a `numpy.ndarray` of dtype `numpy.uint8`.
    """
    warnings.warn(
        "'build_mask' has been deprecated since release 0.16.3. "
        "Use functions from astropy.nddata.bitmask module instead such as "
        "bitfield_to_boolean_mask().",
        DeprecationWarning,
    )
    dqmask = bitfield_to_boolean_mask(
        dqarr,
        bitvalue,
        good_mask_value=1,
        dtype=np.uint8,
        flag_name_map=pixel,
    )
    return dqmask


def calc_gwcs_pixmap(in_wcs, out_wcs, shape=None):
    """
    Generate a pixel map grid using the input and output WCS.

    Parameters
    ----------
    in_wcs : `~astropy.wcs.WCS`
        Input WCS.
    out_wcs : `~astropy.wcs.WCS`
        Output WCS.
    shape : tuple, optional
        Shape of the data. If provided, the bounding box will be calculated
        from the shape. If not provided, the bounding box will be calculated
        from the input WCS.

    Returns
    -------
    pixmap : `~numpy.ndarray`
        The calculated pixel map grid.

    """
    if shape:
        bb = wcs_bbox_from_shape(shape)
        log.debug(f"Bounding box from data shape: {bb}")
    else:
        bb = in_wcs.bounding_box
        log.debug(f"Bounding box from WCS: {in_wcs.bounding_box}")

    grid = gwcs.wcstools.grid_from_bounding_box(bb)
    return np.dstack(reproject(in_wcs, out_wcs)(grid[0], grid[1]))


def reproject(wcs1, wcs2):
    """
    Given two WCSs or transforms return a function which takes pixel
    coordinates in the first WCS or transform and computes them in the second
    one. It performs the forward transformation of ``wcs1`` followed by the
    inverse of ``wcs2``.

    Parameters
    ----------
    wcs1, wcs2 : `astropy.wcs.WCS` or `gwcs.wcs.WCS` or `astropy.modeling.Model`
        WCS objects.

    Returns
    -------
    : func
        Function to compute the transformations.  It takes `(x, y)`
        positions in ``wcs1`` and returns `(x, y)` positions in ``wcs2``.
    """

    if isinstance(wcs1, fitswcs.WCS):
        forward_transform = wcs1.all_pix2world
    elif isinstance(wcs1, gwcs.WCS):
        forward_transform = wcs1.forward_transform
    elif issubclass(wcs1, Model):
        forward_transform = wcs1
    else:
        raise TypeError(
            "Expected input to be astropy.wcs.WCS or gwcs.wcs.WCS "
            "object or astropy.modeling.Model subclass"
        )

    if isinstance(wcs2, fitswcs.WCS):
        backward_transform = wcs2.all_world2pix
    elif isinstance(wcs2, gwcs.WCS):
        backward_transform = wcs2.backward_transform
    elif issubclass(wcs2, Model):
        backward_transform = wcs2.inverse
    else:
        raise TypeError(
            "Expected input to be astropy.wcs.WCS or gwcs.wcs.WCS "
            "object or astropy.modeling.Model subclass"
        )

    def _reproject(x, y):
        sky = forward_transform(x, y)
        flat_sky = []
        for axis in sky:
            flat_sky.append(axis.flatten())
        # Filter out RuntimeWarnings due to computed NaNs in the WCS
        warnings.simplefilter("ignore")
        det = backward_transform(*tuple(flat_sky))
        warnings.resetwarnings()
        det_reshaped = []
        for axis in det:
            det_reshaped.append(axis.reshape(x.shape))
        return tuple(det_reshaped)

    return _reproject


def decode_context(context, x, y):
    """
    Get 0-based indices of input images that contributed to (resampled)
    output pixel with coordinates ``x`` and ``y``.

    Parameters
    ----------
    context: numpy.ndarray
        A 3D `~numpy.ndarray` of integral data type.

    x: int, list of integers, numpy.ndarray of integers
        X-coordinate of pixels to decode (3rd index into the ``context`` array)

    y: int, list of integers, numpy.ndarray of integers
        Y-coordinate of pixels to decode (2nd index into the ``context`` array)

    Returns
    -------

    A list of `numpy.ndarray` objects each containing indices of input images
    that have contributed to an output pixel with coordinates ``x`` and ``y``.
    The length of returned list is equal to the number of input coordinate
    arrays ``x`` and ``y``.

    Examples
    --------

    An example context array for an output image of array shape ``(5, 6)``
    obtained by resampling 80 input images.

    .. code-block:: none

        con = np.array(
            [[[0, 0, 0, 0, 0, 0],
              [0, 0, 0, 36196864, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 537920000, 0, 0, 0]],
             [[0, 0, 0, 0, 0, 0,],
              [0, 0, 0, 67125536, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 163856, 0, 0, 0]],
             [[0, 0, 0, 0, 0, 0],
              [0, 0, 0, 8203, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 32865, 0, 0, 0]]],
            dtype=np.int32
        )
        decode_context(con, [3, 2], [1, 4])
        [array([ 9, 12, 14, 19, 21, 25, 37, 40, 46, 58, 64, 65, 67, 77]),
        array([ 9, 20, 29, 36, 47, 49, 64, 69, 70, 79])]

    """
    if context.ndim != 3:
        raise ValueError("'context' must be a 3D array.")

    x = np.atleast_1d(x)
    y = np.atleast_1d(y)

    if x.size != y.size:
        raise ValueError("Coordinate arrays must have equal length.")

    if x.ndim != 1:
        raise ValueError("Coordinates must be scalars or 1D arrays.")

    if not (np.issubdtype(x.dtype, np.integer) and np.issubdtype(y.dtype, np.integer)):
        raise ValueError("Pixel coordinates must be integer values")

    nbits = 8 * context.dtype.itemsize

    return [
        np.flatnonzero([v & (1 << k) for v in context[:, yi, xi] for k in range(nbits)])
        for xi, yi in zip(x, y)
    ]


def resample_range(data_shape, bbox=None):
    # Find range of input pixels to resample:
    if bbox is None:
        xmin = ymin = 0
        xmax = data_shape[1] - 1
        ymax = data_shape[0] - 1
    else:
        ((x1, x2), (y1, y2)) = bbox
        xmin = max(0, int(x1 + 0.5))
        ymin = max(0, int(y1 + 0.5))
        xmax = min(data_shape[1] - 1, int(x2 + 0.5))
        ymax = min(data_shape[0] - 1, int(y2 + 0.5))

    return xmin, xmax, ymin, ymax



import os

import pytest
import roman_datamodels as rdm
from crds.core.exceptions import CrdsLookupError

from romancal.step import FlatFieldStep
from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_flat_field_image_step(rtdata, ignore_asdf_paths):
    """Test for the flat field step using imaging data."""

    input_data = "r0000101001001001001_0001_wfi01_assignwcs.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test CRDS
    step = FlatFieldStep()
    model = rdm.open(rtdata.input)
    ref_file_path = step.get_reference_file(model, "flat")
    ref_file_name = os.path.split(ref_file_path)[-1]
    assert "roman_wfi_flat" in ref_file_name
    # Test FlatFieldStep
    output = "r0000101001001001001_0001_wfi01_flat.asdf"
    rtdata.output = output
    args = ["romancal.step.FlatFieldStep", rtdata.input]
    RomanStep.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.xfail(reason="Wrong input for the purpose of this test")
@pytest.mark.bigdata
def test_flat_field_grism_step(rtdata, ignore_asdf_paths):
    """Test for the flat field step using grism data. The reference file for
    the grism and prism data should be None, only testing the grism
    case here."""

    input_file = "r0000201001001001001_0001_wfi01_uncal.asdf"
    rtdata.get_data(f"WFI/grism/{input_file}")
    rtdata.input = input_file

    # Test CRDS
    step = FlatFieldStep()
    model = rdm.open(rtdata.input)
    try:
        ref_file_path = step.get_reference_file(model, "flat")
        # Check for a valid reference file
        if ref_file_path != "N/A":
            ref_file_name = os.path.split(ref_file_path)[-1]
        elif ref_file_path == "N/A":
            ref_file_name = ref_file_path
    except CrdsLookupError:
        ref_file_name = None
    assert ref_file_name == "N/A"

    # Test FlatFieldStep
    output = "r0000201001001001001_0001_wfi01_flat.asdf"
    rtdata.output = output
    args = ["romancal.step.FlatFieldStep", rtdata.input]
    RomanStep.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/grism/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.bigdata
@pytest.mark.soctests
def test_flat_field_crds_match_image_step(rtdata, ignore_asdf_paths):
    """DMS79 Test: Testing that different datetimes pull different
    flat files and successfully make level 2 output"""

    # First file
    input_l2_file = "r0000101001001001001_0001_wfi01_assignwcs.asdf"
    rtdata.get_data(f"WFI/image/{input_l2_file}")
    rtdata.input = input_l2_file

    # Test CRDS
    step = FlatFieldStep()
    model = rdm.open(rtdata.input)
    step.log.info(
        "DMS79 MSG: Testing retrieval of best ref file, Success is flat file with"
        " correct use after date"
    )

    step.log.info(f'DMS79 MSG: First data file: {rtdata.input.rsplit("/", 1)[1]}')
    step.log.info(f"DMS79 MSG: Observation date: {model.meta.exposure.start_time}")

    ref_file_path = step.get_reference_file(model, "flat")
    step.log.info(
        f'DMS79 MSG: CRDS matched flat file: {ref_file_path.rsplit("/", 1)[1]}'
    )
    flat = rdm.open(ref_file_path)
    step.log.info(f"DMS79 MSG: flat file UseAfter date: {flat.meta.useafter}")
    step.log.info(
        "DMS79 MSG: UseAfter date before observation date? :"
        f" {(flat.meta.useafter < model.meta.exposure.start_time)}"
    )

    # Test FlatFieldStep
    output = "r0000101001001001001_0001_wfi01_flat.asdf"
    rtdata.output = output
    args = ["romancal.step.FlatFieldStep", rtdata.input]
    step.log.info(
        "DMS79 MSG: Running flat fielding step. The first ERROR is"
        "expected, due to extra CRDS parameters not having been "
        "implemented yet."
    )
    RomanStep.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    step.log.info(
        "DMS79 MSG: Was proper flat fielded "
        "Level 2 data produced? : "
        f"{diff.identical}"
    )
    assert diff.identical, diff.report()

    # This test requires a second file, in order to meet the DMS79 requirement.
    # The test will show that two files with different observation dates match
    #  to separate flat files in CRDS.

    # Second file
    input_file = "r0000101001001001001_0001_wfi01_changetime_assignwcs.asdf"
    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file

    # Test CRDS
    step = FlatFieldStep()
    model = rdm.open(rtdata.input)

    step.log.info(f'DMS79 MSG: Second data file: {rtdata.input.rsplit("/", 1)[1]}')
    step.log.info(f"DMS79 MSG: Observation date: {model.meta.exposure.start_time}")

    ref_file_path_b = step.get_reference_file(model, "flat")
    step.log.info(
        f'DMS79 MSG: CRDS matched flat file: {ref_file_path_b.rsplit("/", 1)[1]}'
    )
    flat = rdm.open(ref_file_path_b)
    step.log.info(f"DMS79 MSG: flat file UseAfter date: {flat.meta.useafter}")
    step.log.info(
        "DMS79 MSG: UseAfter date before observation date? :"
        f" {(flat.meta.useafter < model.meta.exposure.start_time)}"
    )

    # Test FlatFieldStep
    output = "r0000101001001001001_0001_wfi01_changetime_flat.asdf"
    rtdata.output = output
    args = ["romancal.step.FlatFieldStep", rtdata.input]
    step.log.info(
        "DMS79 MSG: Running flat fielding step. The first ERROR is"
        "expected, due to extra CRDS parameters not having been "
        "implemented yet."
    )
    RomanStep.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    step.log.info(
        "DMS79 MSG: Was proper flat fielded "
        "Level 2 data produced? : "
        f"{diff.identical}"
    )
    assert diff.identical, diff.report()

    # Test differing flat matches
    step.log.info(
        "DMS79 MSG REQUIRED TEST: Are the two data files "
        "matched to different flat files? : "
        f'{("/".join(ref_file_path.rsplit("/", 3)[1:]))} != '
        f'{("/".join(ref_file_path_b.rsplit("/", 3)[1:]))}'
    )
    assert "/".join(ref_file_path.rsplit("/", 1)[1:]) != "/".join(
        ref_file_path_b.rsplit("/", 1)[1:]
    )



"""Provide a visual progress bar facility

The actual functionality is provided by the external library `progress`

https://pypi.org/project/progress/

If the module is not available, then stub it out.
"""

import logging

__all__ = ["Bar"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class _BarStub:
    """Stub the Bar functionality"""

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def next(self):
        pass


# Stub the Bar functionality if the actual module does not exist.
try:
    from progress.bar import Bar as _Bar
except ModuleNotFoundError:
    _Bar = _BarStub


def Bar(*args, log_level=logging.INFO, log_cutoff=logging.INFO, **kwargs):
    """Actually use Bar only if logging level is appropriate

    Parameters
    ----------
    *args : *args
        Positional arguments for the `progress.Bar` class

    log_level : int
        The current log level. Only produce a Bar if the `log_level`
        is less than or equal to `log_cutoff`

    log_cutoff : int
        Maximum logging level to allow Bar to be created.

    **kwargs : **kwargs
        Keyword arguments for `progress.Bar` class
    """
    if log_level <= log_cutoff:
        return _Bar(*args, **kwargs)
    return _BarStub(*args, **kwargs)



#! /usr/bin/env python

import roman_datamodels as rdm

from romancal.photom import photom
from romancal.stpipe import RomanStep

__all__ = ["PhotomStep"]


class PhotomStep(RomanStep):
    """
    PhotomStep: Module for loading photometric conversion information from
        reference files and attaching to the input science data model
    """

    class_alias = "photom"

    reference_file_types = ["photom"]

    def process(self, input):
        """Perform the photometric calibration step

        Parameters
        ----------
        input : Roman level 2 image datamodel (wfi_image-1.x.x)
            input roman datamodel

        Returns
        -------
        output_model : Roman level 2 image datamodel (wfi_image-1.x.x)
            output roman datamodel
        """

        if isinstance(input, rdm.DataModel):
            input_model = input
        else:
            input_model = rdm.open(input)

        # Get reference file
        reffile = self.get_reference_file(input_model, "photom")

        # Open the relevant photom reference file as a datamodel
        if reffile is not None and reffile != "N/A":
            # If there is a reference file, perform photom application
            photom_model = rdm.open(reffile)
            self.log.debug(f"Using PHOTOM ref file: {reffile}")

            # Do the correction
            if input_model.meta.exposure.type == "WFI_IMAGE":
                output_model = photom.apply_photom(input_model, photom_model)
                output_model.meta.cal_step.photom = "COMPLETE"
            else:
                self.log.warning("No photometric corrections for spectral data")
                self.log.warning("Photom step will be skipped")
                input_model.meta.cal_step.photom = "SKIPPED"
                input_model.meta.photometry.pixel_area = None
                input_model.meta.photometry.conversion_megajanskys = None
                input_model.meta.photometry.conversion_megajanskys_uncertainty = None
                output_model = input_model
            photom_model.close()

        else:
            # Skip Photom step if no photom file
            self.log.warning("No PHOTOM reference file found")
            self.log.warning("Photom step will be skipped")
            input_model.meta.cal_step.photom = "SKIPPED"
            output_model = input_model

        if self.save_results:
            try:
                self.suffix = "photom"
            except AttributeError:
                self["suffix"] = "photom"

        return output_model



"""
Detect jumps in a science image
"""

import logging
import time

import numpy as np
from roman_datamodels import datamodels as rdd
from roman_datamodels.dqflags import group, pixel
from stcal.jump.jump import detect_jumps

from romancal.stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


__all__ = ["JumpStep"]


class JumpStep(RomanStep):
    """
    JumpStep: Performs CR/jump detection. The 2-point difference method is
    applied.
    """

    class_alias = "jump"

    spec = """
        rejection_threshold = float(default=180.0,min=0) # CR sigma rej thresh
        three_group_rejection_threshold = float(default=185.0,min=0) # CR sigma rej thresh
        four_group_rejection_threshold = float(default=190.0,min=0) # CR sigma rej thresh
        maximum_cores = option('none', 'quarter', 'half', 'all', default='none') # max number of processes to create
        flag_4_neighbors = boolean(default=True) # flag the four perpendicular neighbors of each CR
        max_jump_to_flag_neighbors = float(default=1000) # maximum jump sigma that will trigger neighbor flagging
        min_jump_to_flag_neighbors = float(default=10) # minimum jump sigma that will trigger neighbor flagging
        min_sat_area = float(default=1.0) # minimum required area for the central saturation of snowballs
        min_jump_area = float(default=5.0) # minimum area to trigger large events processing
        expand_factor = float(default=2.0) # The expansion factor for the enclosing circles or ellipses
        use_ellipses = boolean(default=False) # Use an enclosing ellipse rather than a circle for MIRI showers
        sat_required_snowball = boolean(default=True) # Require the center of snowballs to be saturated
        expand_large_events = boolean(default=False) # must be True to trigger snowball and shower flagging
        use_ramp_jump_detection = boolean(default=True) # Use jump detection during ramp fitting
    """  # noqa: E501

    reference_file_types = ["gain", "readnoise"]

    def process(self, input):
        # Open input as a Roman DataModel (single integration; 3D arrays)
        if isinstance(input, rdd.DataModel):
            input_model = input
        else:
            input_model = rdd.open(input)

        # Extract the needed info from the Roman Data Model
        r_data = input_model.data
        r_gdq = input_model.groupdq
        r_pdq = input_model.pixeldq
        r_err = input_model.err
        result = input_model

        # If the ramp fitting jump detection is enabled, then skip this step
        if self.use_ramp_jump_detection:
            result.meta.cal_step.jump = "SKIPPED"
            return result

        # FIXME: since frames_per_group => meta.exposure.nframes has been removed,
        # we need to fix stcal.jump.jump to remove it from there too
        frames_per_group = 1

        # Modify the arrays for input into the 'common' jump (4D)
        data = np.copy(r_data[np.newaxis, :])
        gdq = r_gdq[np.newaxis, :]
        pdq = r_pdq[np.newaxis, :]
        err = np.copy(r_err[np.newaxis, :])

        tstart = time.time()

        # Check for an input model with nresultants<=2
        nresultants = data.shape[1]

        if nresultants <= 2:
            self.log.warning("Cannot apply jump detection as nresultants<=2;")
            self.log.warning("Jump step will be skipped")

            result = input_model

            result.meta.cal_step.jump = "SKIPPED"
            return result

        # Retrieve the parameter values
        rej_thresh = self.rejection_threshold
        three_grp_rej_thresh = self.three_group_rejection_threshold
        four_grp_rej_thresh = self.four_group_rejection_threshold
        max_cores = self.maximum_cores
        max_jump_to_flag_neighbors = self.max_jump_to_flag_neighbors
        min_jump_to_flag_neighbors = self.min_jump_to_flag_neighbors
        flag_4_neighbors = self.flag_4_neighbors
        min_sat_area = self.min_sat_area
        min_jump_area = self.min_jump_area
        expand_factor = self.expand_factor
        use_ellipses = self.use_ellipses
        sat_required_snowball = self.sat_required_snowball
        expand_large_events = self.expand_large_events

        self.log.info("CR rejection threshold = %g sigma", rej_thresh)
        if self.maximum_cores != "none":
            self.log.info("Maximum cores to use = %s", max_cores)

        # Get the gain and readnoise reference files
        # TODO: remove units from gain and RN reference files
        gain_filename = self.get_reference_file(input_model, "gain")
        self.log.info("Using GAIN reference file: %s", gain_filename)
        gain_model = rdd.GainRefModel(gain_filename)
        gain_2d = gain_model.data.value

        readnoise_filename = self.get_reference_file(input_model, "readnoise")
        self.log.info("Using READNOISE reference file: %s", readnoise_filename)
        readnoise_model = rdd.ReadnoiseRefModel(readnoise_filename)
        # This is to clear the WRITEABLE=False flag?
        readnoise_2d = np.copy(readnoise_model.data.value)

        # DG 0810/21:  leave for now; make dqflags changes in a later,
        #              separate PR
        dqflags_d = {}  # Dict of DQ flags
        dqflags_d = {
            "GOOD": group.GOOD,
            "DO_NOT_USE": group.DO_NOT_USE,
            "SATURATED": group.SATURATED,
            "JUMP_DET": group.JUMP_DET,
            "NO_GAIN_VALUE": pixel.NO_GAIN_VALUE,
        }

        gdq, pdq, *_ = detect_jumps(
            frames_per_group,
            data,
            gdq,
            pdq,
            err,
            gain_2d,
            readnoise_2d,
            rej_thresh,
            three_grp_rej_thresh,
            four_grp_rej_thresh,
            max_cores,
            max_jump_to_flag_neighbors,
            min_jump_to_flag_neighbors,
            flag_4_neighbors,
            dqflags_d,
            min_sat_area=min_sat_area,
            min_jump_area=min_jump_area,
            expand_factor=expand_factor,
            use_ellipses=use_ellipses,
            sat_required_snowball=sat_required_snowball,
            expand_large_events=expand_large_events,
        )

        gdq = gdq[0, :, :, :]
        pdq = pdq[0, :, :]
        result.groupdq = gdq
        result.pixeldq = pdq
        gain_model.close()
        readnoise_model.close()
        tstop = time.time()
        self.log.info("The execution time in seconds: %f", tstop - tstart)

        result.meta.cal_step.jump = "COMPLETE"

        if self.save_results:
            try:
                self.suffix = "jump"
            except AttributeError:
                self["suffix"] = "jump"

        return result



"""
Module to calculate the source catalog.
"""

import logging
import warnings
from typing import List

import astropy.units as u
import numpy as np
from astropy.convolution import Gaussian2DKernel
from astropy.coordinates import SkyCoord
from astropy.nddata.utils import NoOverlapError, extract_array
from astropy.stats import SigmaClip, gaussian_fwhm_to_sigma
from astropy.table import QTable, Table
from astropy.utils import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning
from photutils.aperture import CircularAnnulus, CircularAperture, aperture_photometry
from photutils.segmentation import SourceCatalog
from roman_datamodels.datamodels import ImageModel, MosaicModel
from roman_datamodels.dqflags import pixel
from scipy import ndimage
from scipy.spatial import KDTree

from romancal import __version__ as romancal_version
from romancal.lib import psf

from ._wcs_helpers import pixel_scale_angle_at_skycoord

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RomanSourceCatalog:
    """
    Class for the Roman source catalog.

    Parameters
    ----------
    model : `ImageModel` or `MosaicModel`
        The input data model. The image data is assumed to be background
        subtracted.

    segment_image : `~photutils.segmentation.SegmentationImage`
        A 2D segmentation image, with the same shape as the input data,
        where sources are marked by different positive integer values. A
        value of zero is reserved for the background.

    convolved_data : 2D `~numpy.ndarray` or `None`
        The 2D array used to calculate the source centroid and shape
        measurements.

    aperture_params : `dict`
        A dictionary containing the parameters (radii, aperture
        corrections, and background annulus inner and outer radii) used
        to perform aperture photometry.

    ci_star_thresholds : array-like of 2 floats
        The concentration index (CI) thresholds for determining whether
        a source is a star. The first threshold corresponds to the
        concentration index calculated from the smallest and middle
        aperture radii (see ``aperture_params``). The second threshold
        corresponds to the concentration index calculated from the
        middle and largest aperture radii. An object is considered
        extended if both concentration indices are greater than the
        corresponding thresholds, otherwise it is considered a star.

    kernel_fwhm : float
        The full-width at half-maximum (FWHM) of the 2D Gaussian kernel.
        This is needed to calculate the DAOFind sharpness and roundness
        properties (DAOFind uses a special kernel that sums to zero).

    fit_psf : bool
        Whether to fit a PSF model to the sources.

    detection_cat : `None` or `~photutils.segmentation.SourceCatalog`, optional
        A `~photutils.segmentation.SourceCatalog` object for the
        detection image. The segmentation image used to create
        the detection catalog must be the same one input to
        ``segment_image``. If input, then the detection catalog source
        centroids and morphological/shape properties will be returned
        instead of calculating them from the input ``data``. The
        detection catalog centroids and shape properties will also be
        used to perform aperture photometry (i.e., circular and Kron).

    flux_unit : str, optional
        The unit of the flux density. Default is 'uJy'.

    Notes
    -----
    ``model.err`` is assumed to be the total error array corresponding
    to the input science ``model.data`` array. It is assumed to include
    *all* sources of error, including the Poisson error of the sources,
    and have the same shape and units as the science data array.
    """

    def __init__(
        self,
        model,
        segment_img,
        convolved_data,
        aperture_params,
        ci_star_thresholds,
        kernel_fwhm,
        fit_psf,
        detection_cat=None,
        flux_unit="uJy",
    ):

        if not isinstance(model, (ImageModel, MosaicModel)):
            raise ValueError("The input model must be an ImageModel or MosaicModel.")
        self.model = model  # background was previously subtracted

        self.segment_img = segment_img
        self.convolved_data = convolved_data
        self.aperture_params = aperture_params
        self.kernel_sigma = kernel_fwhm * gaussian_fwhm_to_sigma
        self.fit_psf = fit_psf
        self.detection_cat = detection_cat

        if len(ci_star_thresholds) != 2:
            raise ValueError("ci_star_thresholds must contain only 2 items")
        self.ci_star_thresholds = ci_star_thresholds

        self.n_sources = len(self.segment_img.labels)
        self.aperture_ee = self.aperture_params["aperture_ee"]
        self.n_aper = len(self.aperture_ee)
        self.wcs = self.model.meta.wcs
        self.column_desc = {}
        self.meta = {}
        if self.fit_psf:
            # get the necessary columns for the PSF photometry table
            # and its name mapping to the final catalog
            self.psf_photometry_catalog_mapping = (
                self.get_psf_photometry_catalog_colnames_mapping()
            )

        self.flux_unit = flux_unit
        self.l2_conv_factor = self.model.meta.photometry.conversion_megajanskys
        self.sb_to_flux = (1.0 * (u.MJy / u.sr) * self.pixel_area).to(
            u.Unit(self.flux_unit)
        )

        # needed for Kron photometry
        self.segm_sourcecat = None

    @lazyproperty
    def _pixscale_angle(self):
        ysize, xsize = self.model.data.shape
        ycen = (ysize - 1) / 2.0
        xcen = (xsize - 1) / 2.0
        skycoord = self.wcs.pixel_to_world(xcen, ycen)
        _, pixscale, angle = pixel_scale_angle_at_skycoord(skycoord, self.wcs)
        return pixscale, angle

    @lazyproperty
    def _pixel_scale(self):
        """
        The pixel scale in arcseconds at the center of the image.
        """
        return self._pixscale_angle[0]

    @lazyproperty
    def _wcs_angle(self):
        """
        The angle (in degrees) measured counterclockwise from the
        positive x axis to the "North" axis of the celestial coordinate
        system.

        Measured at the center of the image.
        """
        return self._pixscale_angle[1]

    @lazyproperty
    def pixel_area(self):
        """
        The pixel area in steradians.

        If the meta.photometry.pixel_area value is a negative
        placeholder (e.g., -999999 sr), the value is calculated from the
        WCS at the center of the image.
        """
        pixel_area = self.model.meta.photometry.pixel_area * u.sr
        if pixel_area < 0:
            pixel_area = (self._pixel_scale**2).to(u.sr)
        return pixel_area

    def convert_l2_to_sb(self):
        """
        Convert a level-2 image from units of DN/s to MJy/sr (surface
        brightness).
        """
        # the conversion in done in-place to avoid making copies of the data;
        # use a dictionary to set the value to avoid on-the-fly validation
        self.model["data"] *= self.l2_conv_factor
        self.model["err"] *= self.l2_conv_factor
        if self.convolved_data is not None:
            self.convolved_data *= self.l2_conv_factor

    def convert_sb_to_flux_density(self):
        """
        Convert the data and error arrays from MJy/sr (surface
        brightness) to flux density units.

        The arrays are converted in-place to Quantity arrays.

        The flux density unit is defined by self.flux_unit.
        """
        # the conversion in done in-place to avoid making copies of the data;
        # use a dictionary to set the value to avoid on-the-fly validation
        self.model["data"] *= self.sb_to_flux.value
        self.model["data"] <<= self.sb_to_flux.unit
        self.model["err"] *= self.sb_to_flux.value
        self.model["err"] <<= self.sb_to_flux.unit
        if self.convolved_data is not None:
            self.convolved_data *= self.sb_to_flux.value
            self.convolved_data <<= self.sb_to_flux.unit

    def convert_flux_to_abmag(self, flux, flux_err):
        """
        Convert flux (and error) to AB magnitude (and error).

        Parameters
        ----------
        flux, flux_err : `~numpy.ndarray`
            The input flux and error arrays.

        Returns
        -------
        abmag, abmag_err : `~astropy.ndarray`
            The output AB magnitude and error arrays.
        """
        # ignore RunTimeWarning if flux or flux_err contains NaNs
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)

            abmag = flux.to(u.ABmag).value
            abmag_err = 2.5 * np.log10(1.0 + (flux_err / flux))

            # handle negative fluxes
            idx = flux.value < 0
            abmag[idx] = np.nan
            abmag_err[idx] = np.nan

        return abmag, abmag_err

    @lazyproperty
    def segment_colnames(self):
        """
        A dictionary of the output table column names and descriptions
        for the segment catalog.
        """
        desc = {}
        desc["label"] = "Unique source identification label number"
        desc["xcentroid"] = "X pixel value of the source centroid (0 indexed)"
        desc["ycentroid"] = "Y pixel value of the source centroid (0 indexed)"
        desc["sky_centroid"] = " Sky coordinate (ICRS) of the source centroid"
        desc["isophotal_flux"] = "Isophotal flux"
        desc["isophotal_flux_err"] = "Isophotal flux error"
        # isophotal_flux and isophotal_flux_err must be listed before isophotal_abmag
        # TEMP: do not include ABmags
        # desc["isophotal_abmag"] = "Isophotal AB magnitude"
        # desc["isophotal_abmag_err"] = "Isophotal AB magnitude error"
        desc["isophotal_area"] = "Isophotal area"
        desc["kron_flux"] = "Kron flux"
        desc["kron_flux_err"] = "Kron flux error"
        desc["semimajor_sigma"] = (
            "1-sigma standard deviation along the semimajor axis of the 2D Gaussian function that has the same second-order central moments as the source"
        )
        desc["semiminor_sigma"] = (
            "1-sigma standard deviation along the semiminor axis of the 2D Gaussian function that has the same second-order central moments as the source"
        )
        desc["ellipticity"] = (
            "1 minus the ratio of the 1-sigma lengths of the semimajor and semiminor axes"
        )
        desc["orientation"] = (
            "The angle (degrees) between the positive X axis and the major axis (increases counter-clockwise)"
        )
        # orientation must be listed before sky_orientation
        desc["sky_orientation"] = (
            "The position angle (degrees) from North of the major axis"
        )
        desc["sky_bbox_ll"] = (
            "Sky coordinate (ICRS) of the lower-left vertex of the minimal bounding box of the source"
        )
        desc["sky_bbox_ul"] = (
            "Sky coordinate (ICRS) of the upper-left vertex of the minimal bounding box of the source"
        )
        desc["sky_bbox_lr"] = (
            "Sky coordinate (ICRS) of the lower-right vertex of the minimal bounding box of the source"
        )
        desc["sky_bbox_ur"] = (
            "Sky coordinate (ICRS) of the upper-right vertex of the minimal bounding box of the source"
        )

        self.column_desc.update(desc)

        return list(desc.keys())

    def set_segment_properties(self):
        """
        Calculate the segment-based source photometry and morphologies.

        The results are set as dynamic attributes on the class instance.
        """
        if self.detection_cat is not None:
            detection_cat = self.detection_cat.segm_sourcecat
        else:
            detection_cat = None
        segm_cat = SourceCatalog(
            self.model.data,
            self.segment_img,
            convolved_data=self.convolved_data,
            error=self.model.err,
            wcs=self.wcs,
            detection_cat=detection_cat,
        )
        self.segm_sourcecat = segm_cat

        self.meta.update(segm_cat.meta)

        # rename some columns in the output catalog
        prop_names = {}
        prop_names["isophotal_flux"] = "segment_flux"
        prop_names["isophotal_flux_err"] = "segment_fluxerr"
        prop_names["isophotal_area"] = "area"
        prop_names["kron_flux_err"] = "kron_fluxerr"

        # dynamically set the attributes
        for column in self.segment_colnames:
            # use the renamed column name if it exists in prop_names
            prop_name = prop_names.get(column, column)
            try:
                value = getattr(segm_cat, prop_name)
            except AttributeError:
                # isophotal_abmag, isophotal_abmag_err, sky_orientation
                value = getattr(self, prop_name)
            setattr(self, column, value)

    @lazyproperty
    def _xypos(self):
        """
        The (x, y) source positions.

        If a detection catalog is input, then the detection catalog
        centroids are used. Otherwise, the segment catalog centroids are
        used.
        """
        if self.detection_cat is not None:
            xycen = (self.detection_cat.xcentroid, self.detection_cat.ycentroid)
        else:
            xycen = (self.xcentroid, self.ycentroid)
        return np.transpose(xycen)

    @lazyproperty
    def _xypos_aper(self):
        """
        The (x, y) source positions for the circular apertures and
        annuli.

        Non-finite positions are set to a large negative value. At this
        position the aperture will not overlap the data, thus returning
        NaN fluxes and errors.
        """
        xypos = self._xypos.copy()
        nanmask = ~np.isfinite(xypos)
        xypos[nanmask] = -1000.0
        return xypos

    @lazyproperty
    def _xypos_nonfinite_mask(self):
        """
        A 1D boolean mask where `True` values denote sources where
        either the xcentroid or the ycentroid is not finite.
        """
        return ~np.isfinite(self._xypos).all(axis=1)

    @lazyproperty
    def _isophotal_abmag(self):
        """
        The isophotal AB magnitude and error.
        """
        return self.convert_flux_to_abmag(self.isophotal_flux, self.isophotal_flux_err)

    @lazyproperty
    def isophotal_abmag(self):
        """
        The isophotal AB magnitude.
        """
        return self._isophotal_abmag[0]

    @lazyproperty
    def isophotal_abmag_err(self):
        """
        The isophotal AB magnitude error.
        """
        return self._isophotal_abmag[1]

    @lazyproperty
    def sky_orientation(self):
        """
        The orientation of the source major axis as the position angle
        in degrees measured East of North.
        """
        return (180.0 * u.deg) - self._wcs_angle + self.orientation

    def _make_aperture_colnames(self, name):
        """
        Make the aperture column names.

        There are separate columns for the flux/magnitude for each of
        the encircled energies and the total flux/magnitude.

        Parameters
        ----------
        name : {'flux', 'abmag'}
            The name type of the column.

        Returns
        -------
        colnames : list of str
            A list of the output column names.
        """
        colnames = []
        for aper_ee in self.aperture_ee:
            basename = f"aper{aper_ee}_{name}"
            colnames.append(basename)
            colnames.append(f"{basename}_err")
        colnames.extend([f"aper_total_{name}", f"aper_total_{name}_err"])

        return colnames

    def _make_aperture_descriptions(self, name):
        """
        Make aperture column descriptions.

        Parameters
        ----------
        name : {'flux', 'abmag'}
            The name type of the column.

        Returns
        -------
        descriptions : list of str
            A list of the output column descriptions.
        """
        if name == "flux":
            ftype = "Flux"
            ftype2 = "flux"
        elif name == "abmag":
            ftype = ftype2 = "AB magnitude"

        desc = []
        for aper_ee in self.aperture_ee:
            desc.append(
                f"{ftype} within the {aper_ee}% encircled energy circular aperture"
            )
            desc.append(
                f"{ftype} error within the {aper_ee}% encircled energy circular aperture"
            )

        desc.append(
            f"Total aperture-corrected {ftype2} based on the {self.aperture_ee[-1]}% encircled energy circular aperture; should be used only for unresolved sources."
        )
        desc.append(
            f"Total aperture-corrected {ftype2} error based on the {self.aperture_ee[-1]}% encircled energy circular aperture; should be used only for unresolved sources."
        )

        return desc

    @lazyproperty
    def aperture_flux_colnames(self):
        """
        The aperture flux column names.
        """
        return self._make_aperture_colnames("flux")

    @lazyproperty
    def aperture_flux_descriptions(self):
        """
        The aperture flux column descriptions.
        """
        return self._make_aperture_descriptions("flux")

    @lazyproperty
    def aperture_abmag_colnames(self):
        """
        The aperture AB magnitude column names.
        """
        return self._make_aperture_colnames("abmag")

    @lazyproperty
    def aperture_abmag_descriptions(self):
        """
        The aperture AB magnitude column descriptions.
        """
        return self._make_aperture_descriptions("abmag")

    @lazyproperty
    def aperture_colnames(self):
        """
        A dictionary of the output table column names and descriptions
        for the aperture catalog.
        """
        desc = {}
        desc["aper_bkg_flux"] = (
            "The local background value calculated as the sigma-clipped median value in the background annulus aperture"
        )
        desc["aper_bkg_flux_err"] = (
            "The standard error of the sigma-clipped median background value"
        )

        for idx, colname in enumerate(self.aperture_flux_colnames):
            desc[colname] = self.aperture_flux_descriptions[idx]
        # TEMP: do not include ABmags
        # for idx, colname in enumerate(self.aperture_abmag_colnames):
        #     desc[colname] = self.aperture_abmag_descriptions[idx]

        self.column_desc.update(desc)

        return list(desc.keys())

    @lazyproperty
    def _aper_local_background(self):
        """
        Estimate the local background and error using a circular annulus
        aperture.

        The local background is the sigma-clipped median value in the
        annulus.  The background error is the standard error of the
        median, sqrt(pi / 2N) * std.
        """
        bkg_aper = CircularAnnulus(
            self._xypos_aper,
            self.aperture_params["bkg_aperture_inner_radius"],
            self.aperture_params["bkg_aperture_outer_radius"],
        )
        bkg_aper_masks = bkg_aper.to_mask(method="center")
        sigclip = SigmaClip(sigma=3.0)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            warnings.simplefilter("ignore", category=AstropyUserWarning)

            nvalues = []
            bkg_median = []
            bkg_std = []
            for mask in bkg_aper_masks:
                bkg_data = mask.get_values(self.model.data)
                values = sigclip(bkg_data, masked=False)
                nvalues.append(values.size)
                bkg_median.append(np.median(values))
                bkg_std.append(np.std(values))

            nvalues = np.array(nvalues)
            bkg_median = u.Quantity(bkg_median)
            bkg_std = u.Quantity(bkg_std)

            # standard error of the median
            bkg_median_err = np.sqrt(np.pi / (2.0 * nvalues)) * bkg_std

        return bkg_median, bkg_median_err

    @lazyproperty
    def aper_bkg_flux(self):
        """
        The aperture local background flux (per pixel).
        """
        return self._aper_local_background[0]

    @lazyproperty
    def aper_bkg_flux_err(self):
        """
        The aperture local background flux error (per pixel).
        """
        return self._aper_local_background[1]

    def set_aperture_properties(self):
        """
        Calculate the aperture photometry.

        The results are set as dynamic attributes on the class instance.
        """
        apertures = [
            CircularAperture(self._xypos_aper, radius)
            for radius in self.aperture_params["aperture_radii"]
        ]
        aper_phot = aperture_photometry(
            self.model.data, apertures, error=self.model.err
        )

        for i, aperture in enumerate(apertures):
            flux_col = f"aperture_sum_{i}"
            flux_err_col = f"aperture_sum_err_{i}"

            # subtract the local background measured in the annulus
            aper_phot[flux_col] -= self.aper_bkg_flux * aperture.area

            flux = aper_phot[flux_col]
            flux_err = aper_phot[flux_err_col]
            abmag, abmag_err = self.convert_flux_to_abmag(flux, flux_err)

            idx0 = 2 * i
            idx1 = (2 * i) + 1
            setattr(self, self.aperture_flux_colnames[idx0], flux)
            setattr(self, self.aperture_flux_colnames[idx1], flux_err)
            setattr(self, self.aperture_abmag_colnames[idx0], abmag)
            setattr(self, self.aperture_abmag_colnames[idx1], abmag_err)

    @lazyproperty
    def extras_colnames(self):
        """
        A dictionary of the output table column names and descriptions
        for the additional catalog values.
        """
        desc = {}
        for idx, colname in enumerate(self.ci_colnames):
            desc[colname] = self.ci_colname_descriptions[idx]

        desc["flags"] = "Data quality flags"
        desc["is_extended"] = "Flag indicating whether the source is extended"
        desc["sharpness"] = "The DAOFind source sharpness statistic"
        desc["roundness"] = "The DAOFind source roundness statistic"
        desc["nn_label"] = "The label number of the nearest neighbor"
        desc["nn_dist"] = "The distance in pixels to the nearest neighbor"
        self.column_desc.update(desc)

        return list(desc.keys())

    @lazyproperty
    def flags(self):
        """
        Data quality flags.
        """
        xyidx = np.round(self._xypos).astype(int)

        try:
            # L2 images have a dq array
            dqflags = self.model.dq[xyidx[:, 1], xyidx[:, 0]]
            # if dqflags contains the DO_NOT_USE flag, set to DO_NOT_USE
            # (dq=1), otherwise 0
            flags = dqflags & pixel.DO_NOT_USE

        except AttributeError:
            # L3 images
            mask = self.model.weight == 0
            flags = mask[xyidx[:, 1], xyidx[:, 0]].astype(int)

        return flags

    @lazyproperty
    def _ci_ee_indices(self):
        """
        The EE indices for the concentration indices.
        """
        # NOTE: the EE values are always in increasing order
        return (0, 1), (1, 2), (0, 2)

    @lazyproperty
    def ci_colnames(self):
        """
        The column names of the three concentration indices.
        """
        return [
            f"CI_{self.aperture_ee[j]}_{self.aperture_ee[i]}"
            for (i, j) in self._ci_ee_indices
        ]

    @lazyproperty
    def ci_colname_descriptions(self):
        """
        The concentration indices column descriptions.
        """
        return [
            "Concentration index calculated as "
            f"({self.aperture_flux_colnames[2 * j]} / "
            f"{self.aperture_flux_colnames[2 * i]})"
            for (i, j) in self._ci_ee_indices
        ]

    @lazyproperty
    def concentration_indices(self):
        """
        A list of concentration indices, calculated as the flux
        ratios of:

            * the (middle / smallest) aperture flux ratio
              e.g., CI_50_30 = aper50_flux / aper30_flux
            * the (largest / middle) aperture flux ratio
              e.g., CI_70_50 = aper70_flux / aper50_flux
            * the (largest / smallest) aperture flux ratio
              e.g., CI_70_30 = aper70_flux / aper30_flux
        """
        fluxes = [
            (self.aperture_flux_colnames[2 * j], self.aperture_flux_colnames[2 * i])
            for (i, j) in self._ci_ee_indices
        ]
        return [
            getattr(self, flux1).value / getattr(self, flux2).value
            for flux1, flux2 in fluxes
        ]

    def set_ci_properties(self):
        """
        Set the concentration indices as dynamic attributes on the class
        instance.
        """
        for name, value in zip(self.ci_colnames, self.concentration_indices):
            setattr(self, name, value)

    @lazyproperty
    def is_extended(self):
        """
        Boolean indicating whether the source is extended.
        """
        mask1 = self.concentration_indices[0] > self.ci_star_thresholds[0]
        mask2 = self.concentration_indices[1] > self.ci_star_thresholds[1]
        return np.logical_and(mask1, mask2)

    @lazyproperty
    def _daofind_kernel_size(self):
        """
        The DAOFind kernel size (in both x and y dimensions).
        """
        # always odd
        return 2 * int(max(2.0, 1.5 * self.kernel_sigma)) + 1

    @lazyproperty
    def _daofind_kernel_center(self):
        """
        The DAOFind kernel x/y center.
        """
        return (self._daofind_kernel_size - 1) // 2

    @lazyproperty
    def _daofind_kernel_mask(self):
        """
        The DAOFind kernel circular mask.

        NOTE: 1=good pixels, 0=masked pixels
        """
        yy, xx = np.mgrid[0 : self._daofind_kernel_size, 0 : self._daofind_kernel_size]
        radius = np.sqrt(
            (xx - self._daofind_kernel_center) ** 2
            + (yy - self._daofind_kernel_center) ** 2
        )
        return (radius <= max(2.0, 1.5 * self.kernel_sigma)).astype(int)

    @lazyproperty
    def _daofind_kernel(self):
        """
        The DAOFind kernel, a 2D circular Gaussian normalized to have
        zero sum.
        """
        size = self._daofind_kernel_size
        kernel = Gaussian2DKernel(self.kernel_sigma, x_size=size, y_size=size).array
        kernel *= self._daofind_kernel_mask
        kernel /= np.max(kernel)

        # normalize the kernel to zero sum
        npixels = self._daofind_kernel_mask.sum()
        denom = np.sum(kernel**2) - (np.sum(kernel) ** 2 / npixels)
        return ((kernel - (kernel.sum() / npixels)) / denom) * self._daofind_kernel_mask

    @lazyproperty
    def _daofind_convolved_data(self):
        """
        The DAOFind convolved data.
        """
        return ndimage.convolve(
            self.model.data.value, self._daofind_kernel, mode="constant", cval=0.0
        )

    @lazyproperty
    def _daofind_cutout(self):
        """
        3D array containing 2D cutouts centered on each source from the
        input data.

        The cutout size always matches the size of the DAOFind kernel,
        which has odd dimensions.
        """
        cutout = []
        for xcen, ycen in zip(*np.transpose(self._xypos_aper)):
            try:
                cutout_ = extract_array(
                    self.model.data,
                    self._daofind_kernel.shape,
                    (ycen, xcen),
                    fill_value=0.0,
                )
            except NoOverlapError:
                cutout_ = np.zeros(self._daofind_kernel.shape)
            cutout.append(cutout_)

        return np.array(cutout)  # all cutouts are the same size

    @lazyproperty
    def _daofind_cutout_conv(self):
        """
        3D array containing 2D cutouts centered on each source from the
        DAOFind convolved data.

        The cutout size always matches the size of the DAOFind kernel,
        which has odd dimensions.
        """
        cutout = []
        for xcen, ycen in zip(*np.transpose(self._xypos_aper)):
            try:
                cutout_ = extract_array(
                    self._daofind_convolved_data,
                    self._daofind_kernel.shape,
                    (ycen, xcen),
                    fill_value=0.0,
                )
            except NoOverlapError:
                cutout_ = np.zeros(self._daofind_kernel.shape)
            cutout.append(cutout_)

        return np.array(cutout)  # all cutouts are the same size

    @lazyproperty
    def sharpness(self):
        """
        The DAOFind source sharpness statistic.

        The sharpness statistic measures the ratio of the difference
        between the height of the central pixel and the mean of the
        surrounding non-bad pixels to the height of the best fitting
        Gaussian function at that point.

        Stars generally have a ``sharpness`` between 0.2 and 1.0.
        """
        npixels = self._daofind_kernel_mask.sum() - 1  # exclude the peak pixel
        data_masked = self._daofind_cutout * self._daofind_kernel_mask
        data_peak = self._daofind_cutout[
            :, self._daofind_kernel_center, self._daofind_kernel_center
        ]
        conv_peak = self._daofind_cutout_conv[
            :, self._daofind_kernel_center, self._daofind_kernel_center
        ]

        data_mean = (np.sum(data_masked, axis=(1, 2)) - data_peak) / npixels

        with warnings.catch_warnings():
            # ignore 0 / 0 for non-finite xypos
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return (data_peak - data_mean) / conv_peak

    @lazyproperty
    def roundness(self):
        """
        The DAOFind source roundness statistic based on symmetry.

        The roundness characteristic computes the ratio of a measure of
        the bilateral symmetry of the object to a measure of the
        four-fold symmetry of the object.

        "Round" objects have a ``roundness`` close to 0, generally
        between -1 and 1.
        """
        # set the central (peak) pixel to zero
        cutout = self._daofind_cutout_conv.copy()
        cutout[:, self._daofind_kernel_center, self._daofind_kernel_center] = 0.0

        # calculate the four roundness quadrants
        quad1 = cutout[
            :, 0 : self._daofind_kernel_center + 1, self._daofind_kernel_center + 1 :
        ]
        quad2 = cutout[
            :, 0 : self._daofind_kernel_center, 0 : self._daofind_kernel_center + 1
        ]
        quad3 = cutout[
            :, self._daofind_kernel_center :, 0 : self._daofind_kernel_center
        ]
        quad4 = cutout[
            :, self._daofind_kernel_center + 1 :, self._daofind_kernel_center :
        ]

        axis = (1, 2)
        sum2 = (
            -quad1.sum(axis=axis)
            + quad2.sum(axis=axis)
            - quad3.sum(axis=axis)
            + quad4.sum(axis=axis)
        )
        sum2[sum2 == 0] = 0.0

        sum4 = np.abs(cutout).sum(axis=axis)
        sum4[sum4 == 0] = np.nan

        with warnings.catch_warnings():
            # ignore 0 / 0 for non-finite xypos
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return 2.0 * sum2 / sum4

    @lazyproperty
    def _kdtree_query(self):
        """
        The distance in pixels to the nearest neighbor and its index.
        """
        if self.n_sources == 1:
            return [np.nan], [np.nan]

        # non-finite xypos causes memory errors on linux, but not MacOS
        tree = KDTree(self._xypos_aper)
        qdist, qidx = tree.query(self._xypos_aper, k=[2])
        return np.transpose(qdist)[0], np.transpose(qidx)[0]

    @lazyproperty
    def nn_label(self):
        """
        The label number of the nearest neighbor.

        A label value of -1 is returned if there is only one detected
        source and for sources with a non-finite xcentroid or ycentroid.
        """
        if self.n_sources == 1:
            return -1

        nn_label = self.label[self._kdtree_query[1]]
        # assign a label of -1 for non-finite xypos
        nn_label[self._xypos_nonfinite_mask] = -1

        return nn_label

    @lazyproperty
    def nn_dist(self):
        """
        The distance in pixels to the nearest neighbor.
        """
        nn_dist = self._kdtree_query[0]
        if self.n_sources == 1:
            # NaN if only one detected source
            return nn_dist * u.pixel

        # assign a distance of np.nan for non-finite xypos
        nn_dist[self._xypos_nonfinite_mask] = np.nan
        return nn_dist * u.pixel

    @lazyproperty
    def aper_total_flux(self):
        """
        The aperture-corrected total flux for sources, based on the flux
        in largest aperture.

        The aperture-corrected total flux should be used only for
        unresolved sources.
        """
        idx = self.n_aper - 1  # use apcorr for the largest EE (largest radius)
        flux = self.aperture_params["aperture_corrections"][idx] * getattr(
            self, self.aperture_flux_colnames[idx * 2]
        )
        return flux

    @lazyproperty
    def aper_total_flux_err(self):
        """
        The aperture-corrected total flux error for sources,
        based on the flux in largest aperture.

        The aperture-corrected total flux error should be used only for
        unresolved sources.
        """
        idx = self.n_aper - 1  # use apcorr for the largest EE (largest radius)
        flux_err = self.aperture_params["aperture_corrections"][idx] * getattr(
            self, self.aperture_flux_colnames[idx * 2 + 1]
        )
        return flux_err

    @lazyproperty
    def _abmag_total(self):
        """
        The total AB magnitude and error.
        """
        return self.convert_flux_to_abmag(
            self.aper_total_flux, self.aper_total_flux_err
        )

    @lazyproperty
    def aper_total_abmag(self):
        """
        The aperture-corrected total AB magnitude.

        The aperture-corrected total magnitude should be used only for
        unresolved sources.
        """
        return self._abmag_total[0]

    @lazyproperty
    def aper_total_abmag_err(self):
        """
        The aperture-corrected total AB magnitude error.

        The aperture-corrected total magnitude error should be used only
        for unresolved sources.
        """
        return self._abmag_total[1]

    @lazyproperty
    def colnames(self):
        """
        The column name order for the final source catalog.
        """
        if self.detection_cat is None:
            colnames = self.segment_colnames[:4]
            colnames.extend(self.aperture_colnames)
            colnames.extend(self.extras_colnames)
            colnames.extend(self.segment_colnames[4:])
            if self.fit_psf:
                colnames.extend(self.psf_photometry_colnames)

        else:
            # NOTE: label is need to join the tables
            colnames = [
                "label",
                "isophotal_flux",
                "isophotal_flux_err",
                "kron_flux",
                "kron_flux_err",
            ]
            colnames.extend(self.aperture_colnames)
            if self.fit_psf:
                colnames.extend(self.psf_photometry_colnames)

        return colnames

    def _update_metadata(self):
        """
        Update the metadata dictionary with the package version
        information and aperture parameters.
        """
        ver_key = "versions"
        if "version" in self.meta.keys():
            # depends on photutils version
            self.meta[ver_key] = self.meta.pop("version")

        ver_dict = self.meta.get("versions", None)
        if ver_dict is not None:
            ver_dict["romancal"] = romancal_version
            packages = [
                "Python",
                "numpy",
                "scipy",
                "astropy",
                "photutils",
                "gwcs",
                "romancal",
            ]
            ver_dict = {key: ver_dict[key] for key in packages if key in ver_dict}
            self.meta[ver_key] = ver_dict

        self.meta["aperture_params"] = self.aperture_params

    def _split_skycoord(self, table):
        """
        Split SkyCoord columns into separate RA and Dec columns.

        Parameters
        ----------
        table : `~astropy.table.Table`
            The input table.

        Returns
        -------
        table : `~astropy.table.Table`
            The output table with separate RA and Dec columns.
        """
        for colname in table.colnames:
            if isinstance(table[colname], SkyCoord):
                idx = table.colnames.index(colname)
                ra = table[colname].ra
                dec = table[colname].dec
                ra_colname = colname.replace("sky", "ra")
                dec_colname = colname.replace("sky", "dec")
                table.remove_column(colname)
                table.add_columns(
                    [ra, dec], names=[ra_colname, dec_colname], indexes=[idx, idx]
                )
                desc = self.column_desc[colname]
                ra_desc = desc.replace("Sky coordinate", "Right ascension")
                dec_desc = desc.replace("Sky coordinate", "Declination")
                table[ra_colname].info.description = ra_desc
                table[dec_colname].info.description = dec_desc

        return table

    @staticmethod
    def get_psf_photometry_catalog_colnames_mapping() -> List:
        """
        Set the mapping between the PSF photometry table column names
        and the final catalog column names.

        Returns
        -------
            List of dictionaries containing old column names, new column names,
            and descriptions for PSF photometry catalog.
        """

        return [
            {
                "old_name": "flags",
                "new_name": "flag_psf",
                "desc": "Data quality flags",
            },
            {
                "old_name": "x_fit",
                "new_name": "x_psf",
                "desc": "X coordinate as determined by PSF fitting",
            },
            {
                "old_name": "x_err",
                "new_name": "x_psf_err",
                "desc": "Error on X coordinate of PSF fitting",
            },
            {
                "old_name": "y_fit",
                "new_name": "y_psf",
                "desc": "Y coordinate as determined by PSF fitting",
            },
            {
                "old_name": "y_err",
                "new_name": "y_psf_err",
                "desc": "Error on Y coordinate of PSF fitting",
            },
            {
                "old_name": "flux_fit",
                "new_name": "flux_psf",
                "desc": "Source flux as determined by PSF photometry",
            },
            {
                "old_name": "flux_err",
                "new_name": "flux_psf_err",
                "desc": "Source flux error as determined by PSF photometry",
            },
        ]

    @lazyproperty
    def psf_photometry_colnames(self) -> List:
        """
        Update and return column descriptions for PSF photometry results.

        This method updates the column descriptions with PSF photometry-related information
        and returns a list of column names.

        Returns
        -------
            List of column names for PSF photometry results.
        """
        desc = {x["new_name"]: x["desc"] for x in self.psf_photometry_catalog_mapping}

        self.column_desc.update(desc)

        return list(desc.keys())

    def do_psf_photometry(self) -> None:
        """
        Perform PSF photometry by fitting PSF models to detected sources for refined astrometry.

        This method constructs a gridded PSF model based on instrument and detector information.
        It then fits the PSF model to the image model's sources to improve astrometric precision.

        """
        log.info("Constructing a gridded PSF model.")
        if hasattr(self.model.meta, "instrument"):
            # ImageModel (L2 datamodel)
            filt = self.model.meta.instrument.optical_element
            detector = self.model.meta.instrument.detector.replace("WFI", "SCA")
        else:
            # MosaicModel (L3 datamodel)
            filt = self.model.meta.basic.optical_element
            detector = "SCA02"
        # prefix of the temporary FITS file that will contain the gridded PSF model
        gridded_psf_model, _ = psf.create_gridded_psf_model(
            filt=filt,
            detector=detector,
        )

        log.info("Fitting a PSF model to sources for improved astrometric precision.")
        xinit, yinit = np.transpose(self._xypos)
        psf_photometry_table, photometry = psf.fit_psf_to_image_model(
            image_model=self.model,
            psf_model=gridded_psf_model,
            x_init=xinit,
            y_init=yinit,
            exclude_out_of_bounds=True,
        )

        # mapping between the columns of the PSF photometry table
        # and the name that will be used in the final catalog
        old_name_to_new_name_mapping = {
            x["old_name"]: x["new_name"] for x in self.psf_photometry_catalog_mapping
        }

        # append PSF results to the class instance with the proper column name
        for old_name, new_name in old_name_to_new_name_mapping.items():
            setattr(self, new_name, psf_photometry_table[old_name])

    @lazyproperty
    def catalog(self):
        """
        The final source catalog.
        """
        # convert L2 data units back to MJy/sr
        if isinstance(self.model, ImageModel):
            self.convert_l2_to_sb()

        self.convert_sb_to_flux_density()
        self.set_segment_properties()
        self.set_aperture_properties()
        self.set_ci_properties()
        if self.fit_psf:
            self.do_psf_photometry()

        catalog = QTable()
        for column in self.colnames:
            catalog[column] = getattr(self, column)
            catalog[column].info.description = self.column_desc[column]
        self._update_metadata()
        catalog.meta.update(self.meta)

        # convert QTable to Table to avoid having Quantity columns
        catalog = Table(catalog)

        # split SkyCoord columns into separate RA and Dec columns
        catalog = self._split_skycoord(catalog)

        return catalog



"""
Module for the multiband source catalog step.
"""

import logging

import numpy as np
from astropy.table import Table, join
from roman_datamodels import datamodels, maker_utils

from romancal.datamodels import ModelLibrary
from romancal.multiband_catalog.background import subtract_background_library
from romancal.multiband_catalog.detection_image import make_detection_image
from romancal.multiband_catalog.utils import prefix_colnames, remove_columns
from romancal.source_catalog.background import RomanBackground
from romancal.source_catalog.detection import make_segmentation_image
from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.stpipe import RomanStep

__all__ = ["MultibandCatalogStep"]

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class MultibandCatalogStep(RomanStep):
    """
    Create a multiband catalog of sources including photometry and basic
    shape measurements.

    Parameters
    -----------
    input : str or `~romancal.datamodels.ModelLibrary`
        Path to an ASDF file or a `~romancal.datamodels.ModelLibrary`
        that contains `~roman_datamodels.datamodels.MosaicImageModel`
        models.
    """

    class_alias = "multiband_catalog"
    reference_file_types = []

    spec = """
        bkg_boxsize = integer(default=100)   # background mesh box size in pixels
        kernel_fwhms = float_list(default=None)  # Gaussian kernel FWHM in pixels
        snr_threshold = float(default=3.0)    # per-pixel SNR threshold above the bkg
        npixels = integer(default=25)         # min number of pixels in source
        deblend = boolean(default=False)      # deblend sources?
        aperture_ee1 = integer(default=30)    # aperture encircled energy 1
        aperture_ee2 = integer(default=50)    # aperture encircled energy 2
        aperture_ee3 = integer(default=70)    # aperture encircled energy 3
        ci1_star_threshold = float(default=2.0)  # CI 1 star threshold
        ci2_star_threshold = float(default=1.8)  # CI 2 star threshold
        suffix = string(default='cat')        # Default suffix for output files
        fit_psf = boolean(default=True)      # fit source PSFs for accurate astrometry?
    """

    def process(self, library):
        # All input MosaicImages in the ModelLibrary are assumed to have
        # the same shape and be pixel aligned.
        if isinstance(library, str):
            library = ModelLibrary(library)
        if not isinstance(library, ModelLibrary):
            raise TypeError("library input must be a ModelLibrary object")

        cat_model = datamodels.MosaicSourceCatalogModel
        source_catalog_model = maker_utils.mk_datamodel(cat_model)
        try:
            source_catalog_model.meta.filename = library.asn["products"][0]["name"]
        except (AttributeError, KeyError):
            source_catalog_model.meta.filename = "multiband_catalog"
        if self.output_file is None:
            self.output_file = source_catalog_model.meta.filename

        # TODO: sensible defaults
        # TODO: redefine in terms of intrinsic FWHM
        if self.kernel_fwhms is None:
            self.kernel_fwhms = [2.0, 20.0]

        # define the aperture parameters for the source catalog
        # based on the input encircled energy fractions
        # TODO: define these values in RomanSourceCatalog
        aperture_ee = (self.aperture_ee1, self.aperture_ee2, self.aperture_ee3)
        aperture_params = {
            "aperture_radii": np.array((1.0, 2.0, 3.0)),
            "aperture_corrections": np.array((1.0, 1.0, 1.0)),
            "aperture_ee": aperture_ee,
            "bkg_aperture_inner_radius": 5.0,
            "bkg_aperture_outer_radius": 10.0,
        }

        library = subtract_background_library(library, self.bkg_boxsize)

        # TODO: where to save the det_img and det_err?
        det_img, det_err = make_detection_image(library, self.kernel_fwhms)

        # estimate background rms from detection image to calculate a
        # threshold for source detection
        mask = np.isnan(det_img)
        coverage_mask = np.isnan(det_err)
        bkg = RomanBackground(
            det_img,
            box_size=self.bkg_boxsize,
            mask=mask,
            coverage_mask=coverage_mask,
        )
        bkg_rms = bkg.background_rms

        segment_img = make_segmentation_image(
            det_img,
            snr_threshold=self.snr_threshold,
            npixels=self.npixels,
            bkg_rms=bkg_rms,
            deblend=self.deblend,
            mask=coverage_mask,
        )

        if segment_img is None:  # no sources found
            source_catalog_model.source_catalog = Table()
        else:
            ci_star_thresholds = (
                self.ci1_star_threshold,
                self.ci2_star_threshold,
            )

            # this is needed for the DAOStarFinder sharpness and roundness
            # properties
            # TODO: measure on a secondary detection image with minimal
            # smoothing; same for basic shape parameters
            star_kernel_fwhm = np.min(self.kernel_fwhms)  # ??

            det_model = maker_utils.mk_datamodel(datamodels.MosaicModel)
            det_model.data = det_img
            det_model.err = det_err

            # TODO: this is a temporary solution to get model attributes
            # currently needed in RomanSourceCatalog
            with library:
                model = library.borrow(0)
                det_model.weight = model.weight
                det_model.meta = model.meta
                library.shelve(model, modify=False)

            det_catobj = RomanSourceCatalog(
                det_model,
                segment_img,
                det_img,
                aperture_params,
                ci_star_thresholds,
                star_kernel_fwhm,
                self.fit_psf,
                detection_cat=None,
            )
            det_cat = remove_columns(det_catobj.catalog)

            # loop over each image
            with library:
                for model in library:
                    catobj = RomanSourceCatalog(
                        model,
                        segment_img,
                        None,
                        aperture_params,
                        ci_star_thresholds,
                        star_kernel_fwhm,
                        self.fit_psf,
                        detection_cat=det_catobj,
                    )

                    filter_name = model.meta.basic.optical_element
                    prefix = f"{filter_name}_"
                    cat = prefix_colnames(catobj.catalog, prefix=prefix)
                    # TODO: what metadata do we want to keep, if any,
                    # from the filter catalogs?
                    cat.meta = None  # temporary
                    # outer join prevents empty table if any columns have
                    # the same name but different values (e.g., repeated
                    # filter names)
                    det_cat = join(det_cat, cat, keys="label", join_type="outer")
                    library.shelve(model, modify=False)

            # put the resulting catalog in the model
            source_catalog_model.source_catalog = det_cat

        # explicitly save the segmentation image;
        # the multiband catalog is automatically saved by stpipe
        self.save_segm(segment_img, source_catalog_model)

        return source_catalog_model

    def save_segm(self, segment_img, source_catalog_model):
        """
        Save the segmentation image.
        """
        output_filename = (
            self.output_file
            if self.output_file is not None
            else source_catalog_model.meta.filename
        )

        seg_model = datamodels.MosaicSegmentationMapModel
        segmentation_model = maker_utils.mk_datamodel(seg_model)
        for key in segmentation_model.meta.keys():
            segmentation_model.meta[key] = source_catalog_model.meta[key]

        if segment_img is not None:
            segmentation_model.data = segment_img.data.astype(np.uint32)
            self.save_model(
                segmentation_model,
                output_file=output_filename,
                suffix="segm",
                force=True,
            )



"""Helpers for tests."""

# ruff: noqa
import os
import re
from collections import namedtuple
from contextlib import contextmanager
from glob import glob
from tempfile import TemporaryDirectory

import pytest
from astropy.table import Table, vstack

from romancal.associations import AssociationPool, AssociationRegistry, generate
from romancal.associations.lib.counter import Counter
from romancal.associations.lib.utilities import is_iterable


# Define how to setup initial conditions with pools.
class PoolParams(
    namedtuple(
        "PoolParams",
        ["path", "n_asns", "n_orphaned", "candidates", "valid_suffixes", "kwargs"],
    )
):
    def __new__(
        cls,
        path="",
        n_asns=0,
        n_orphaned=0,
        candidates=None,
        valid_suffixes=None,
        kwargs=None,
    ):
        if not kwargs:
            kwargs = {}
        if candidates is None:
            candidates = []
            if valid_suffixes is None:
                valid_suffixes = ["cal", "calints", "cat"]
        return super().__new__(
            cls, path, n_asns, n_orphaned, candidates, valid_suffixes, kwargs
        )


# Basic Pool/Rule test class
class BasePoolRule:
    # Define the pools and testing parameters related to them.
    # Each entry is a tuple starting with the path of the pool.
    pools = []

    # Define the rules that SHOULD be present.
    # Each entry is the class name of the rule.
    valid_rules = []

    def test_rules_exist(self):
        rules = registry_level2_only()
        assert len(rules) >= len(self.valid_rules)
        rule_names = get_rule_names(rules)
        for rule in self.valid_rules:
            assert rule in rule_names

    def test_run_generate(self):
        rules = registry_level2_only()
        for ppars in self.pools:
            pool = combine_pools(ppars.path, **ppars.kwargs)
            asns = generate(pool, rules)
            assert (
                len(asns) == ppars.n_asns
            ), ppars.path + ": n_asns not expected {} {}".format(
                len(asns), ppars.n_asns
            )
            for asn, candidates in zip(asns, ppars.candidates):
                assert set(asn.candidates) == set(candidates)
            file_regex = re.compile(r".+_(?P<suffix>.+)\..+")
            for asn in asns:
                for product in asn["products"]:
                    for member in product["members"]:
                        if member["exptype"] == "science":
                            match = file_regex.match(member["expname"])
                            assert (
                                match is not None
                            ), ppars.path + ": No suffix match for {}".format(
                                member["expname"]
                            )
                            assert (
                                match.groupdict()["suffix"] in ppars.valid_suffixes
                            ), ppars.path + ": Suffix {} not valid".format(
                                match.groupdict()["suffix"]
                            )


def make_megapool():
    """Combine the individual test pools into one

    Notes
    -----
    This is meant to be run in the source tree in
    the folder the test pools reside in. The package
    does need to have been installed.
    `python -c 'import romancal.associations.tests.helpers as helpers; helpers.make_megapool()'` # noqa: E501
    """
    pool_files = glob("pool_*.csv")
    pool_files.sort()
    pool = combine_pools(pool_files)
    pool.write("mega_pool.csv", format="ascii", delimiter="|", overwrite=True)


# Basic utilities.
def check_in_list(element, alist):
    assert element in alist


def check_not_in_list(element, alist):
    assert element not in alist


def check_equal(left, right):
    assert left == right


def not_none(value):
    assert value is not None


def t_path(partial_path):
    """Construction the full path for test files"""
    test_dir = os.path.dirname(__file__)
    return os.path.join(test_dir, partial_path)


def combine_pools(pools, **pool_kwargs):
    """Combine pools into a single pool

    Parameters
    ----------
    pools: str, astropy.table.Table, [str|Table, ...]
        The pools to combine. Either a singleton is
        passed or and iterable can be passed.
        The entries themselves can be either a file path
        or an astropy.table.Table-like object.

    pool_kwargs: dict
        Other keyword arguments to pass to AssociationPool.read

    Returns
    -------
    AssociationPool|astropy.table.Table
        The combined pool
    """
    if not is_iterable(pools):
        pools = [pools]
    just_pools = []
    for pool in pools:
        if not isinstance(pool, Table):
            pool = AssociationPool.read(pool, **pool_kwargs)
        just_pools.append(pool)
    if len(just_pools) > 1:
        mega_pool = vstack(just_pools, metadata_conflicts="silent")
    else:
        mega_pool = just_pools[0].copy(copy_data=True)

    # Replace OBS_NUM and ASN_CANDIDATE_ID with actual numbers, if
    # necessary
    expnum = Counter(start=0)
    obsnum = Counter(start=0)
    acid = Counter(start=999)
    local_env = locals()
    global_env = globals()
    for row in mega_pool:
        mega_pool[row.index] = [
            parse_value(v, global_env=global_env, local_env=local_env) for v in row
        ]

    return mega_pool


def parse_value(v, global_env=None, local_env=None):
    """Evaluate if indicated"""
    if global_env is None:
        global_env = globals()
    if local_env is None:
        local_env = locals()

    result = v
    try:
        m = re.match("@!(.+)", v)
    except TypeError:
        pass
    else:
        if m:
            result = eval(m.group(1), global_env, local_env)
    return result


def fmt_cand(candidate_list):
    """Format the candidate field

    Parameters
    ----------
    candidate_list: iterator
        An iterator with each element a 2-tuple of:
            cid: int
                Candidate ID
            ctype: str
                Candidate type

    Returns
    -------
    candidate_list_field: str
        A string of the list of candidates, with any evaluation
        performed.
    """
    evaled_list = []
    for cid, ctype in candidate_list:
        if isinstance(cid, int):
            if ctype == "observation" and cid < 1000:
                cid_format = "o{:0>3d}"
            elif cid >= 1000 and cid < 3000:
                cid_format = "c{:0>4d}"
            else:
                cid_format = "r{:0>4d}"
        else:
            cid_format = cid

        cid_str = cid_format.format(cid)
        evaled_list.append((cid_str, ctype))

    return str(evaled_list)


def fmt_fname(expnum):
    """Format the filename"""
    return f"jw_{expnum:0>5d}_uncal.fits"


def generate_params(request):
    """Simple param reflection for pytest.fixtures"""
    return request.param


def func_fixture(f, **kwargs):
    """Create a true decorator for pytest.fixture

    Parameters
    ----------
    f: func
        The function.

    kwargs: dict
        Keyword arguments to pass to pytest.fixture
    """

    @pytest.fixture(**kwargs)
    def duped(request, *duped_args, **duped_kwargs):
        return f(request, *duped_args, **duped_kwargs)

    return duped


@contextmanager
def mkstemp_pool_file(pools, **pool_kwargs):
    """Make an actual pool file"""
    pool = combine_pools(pools, **pool_kwargs)
    with TemporaryDirectory() as path:
        pool_path = os.path.join(path, "pool")
        pool.write(
            pool_path,
            format="ascii",
            delimiter="|",
        )
        yield pool_path


def generate_pool_paths(request):
    """Fixture to create temporary files for pools"""
    pool_file = t_path(request.param)
    with mkstemp_pool_file(pool_file) as pool_path:
        yield pool_path


def get_rule_names(rules):
    """Return rules names found in a registry

    Parameters
    ----------
    rules: AssociationRegistry
        The registry to look through

    Returns
    -------
    rule_names: list
        The list of rule names
    """
    return [rule._asn_rule() for rule_name, rule in rules.items()]


def level2_rule_path():
    """Return the path to the level 2 rules"""
    return t_path("../lib/rules_level2.py")


def registry_level2_only(global_constraints=None):
    """Get registry with only Level2 rules"""
    return AssociationRegistry(
        definition_files=[level2_rule_path()],
        include_default=False,
        global_constraints=global_constraints,
    )



import json

import pytest
import roman_datamodels.datamodels as dm
from roman_datamodels.maker_utils import mk_level2_image

from romancal.associations import load_asn
from romancal.associations.asn_from_list import asn_from_list
from romancal.datamodels.library import ModelLibrary

# for the example association, set 2 different observation numbers
# so the association will have 2 groups (since all other group_id
# determining meta is the same, see `example_asn_path`)
_OBSERVATION_NUMBERS = [1, 1, 2]
_N_MODELS = len(_OBSERVATION_NUMBERS)
_N_GROUPS = len(set(_OBSERVATION_NUMBERS))
_PRODUCT_NAME = "foo_out"


@pytest.fixture
def example_asn_path(tmp_path):
    """
    Fixture that creates a simple association, saves it (and the models)
    to disk, and returns the path of the saved association
    """
    fns = []
    for i in range(_N_MODELS):
        m = dm.ImageModel(mk_level2_image(shape=(2, 2)))
        m.meta.observation.program = 1
        m.meta.observation.observation = _OBSERVATION_NUMBERS[i]
        m.meta.observation.visit = 1
        m.meta.observation.visit_file_group = 1
        m.meta.observation.visit_file_sequence = 1
        m.meta.observation.visit_file_activity = "01"
        m.meta.observation.exposure = 1
        base_fn = f"{i}.asdf"
        m.meta.filename = base_fn
        m.save(str(tmp_path / base_fn))
        fns.append(base_fn)
    asn = asn_from_list(fns, product_name=_PRODUCT_NAME)
    base_fn, contents = asn.dump(format="json")
    asn_filename = tmp_path / base_fn
    with open(asn_filename, "w") as f:
        f.write(contents)
    return asn_filename


@pytest.fixture
def example_library(example_asn_path):
    """
    Fixture that builds off of `example_asn_path` and returns a
    library created from the association with default options
    """
    return ModelLibrary(example_asn_path)


def _set_custom_member_attr(example_asn_path, member_index, attr, value):
    """
    Helper function to modify the association at `example_asn_path`
    by adding an attribute `attr` to the member list (at index
    `member_index`) with value `value`. This is used to modify
    the `group_id` or `exptype` of a certain member for some tests.
    """
    with open(example_asn_path) as f:
        asn_data = load_asn(f)
    asn_data["products"][0]["members"][member_index][attr] = value
    with open(example_asn_path, "w") as f:
        json.dump(asn_data, f)


def test_assign_member(example_asn_path):
    exptypes = ["science"] * _N_MODELS
    _set_custom_member_attr(example_asn_path, 1, "exptype", "background")
    exptypes[1] = "background"

    def get_exptype(model, index):
        return model.meta.exptype.lower()

    library = ModelLibrary(example_asn_path)

    assert list(library.map_function(get_exptype)) == exptypes


@pytest.mark.parametrize("attr", ["group_names", "group_indices"])
def test_group_with_no_datamodels_open(example_asn_path, attr, monkeypatch):
    """
    Test that the "grouping" methods do not call datamodels.open
    """

    # patch datamodels.open to always raise an exception
    # this will serve as a smoke test to see if any of the attribute
    # accesses (or instance creation) attempts to open models
    def no_open(*args, **kwargs):
        raise Exception()

    monkeypatch.setattr(dm, "open", no_open)

    # use example_asn_path here to make the instance after we've patched
    # datamodels.open
    library = ModelLibrary(example_asn_path)
    getattr(library, attr)


def test_asn_data(example_library):
    """
    Test that `asn` returns the association information
    """
    assert example_library.asn["products"][0]["name"] == _PRODUCT_NAME



"""
Roman step for sky matching.
"""

import logging
from copy import deepcopy

import numpy as np
from astropy.nddata.bitmask import bitfield_to_boolean_mask, interpret_bit_flags
from roman_datamodels import maker_utils
from roman_datamodels.dqflags import pixel

from romancal.datamodels import ModelLibrary
from romancal.stpipe import RomanStep

from .skyimage import SkyImage
from .skymatch import match
from .skystatistics import SkyStats

__all__ = ["SkyMatchStep"]


class SkyMatchStep(RomanStep):
    """
    SkyMatchStep: Subtraction or equalization of sky background in science images.
    """

    class_alias = "skymatch"

    spec = """
        # General sky matching parameters:
        skymethod = option('local', 'global', 'match', 'global+match', default='match') # sky computation method
        match_down = boolean(default=True) # adjust sky to lowest measured value?
        subtract = boolean(default=False) # subtract computed sky from image data?

        # Image's bounding polygon parameters:
        stepsize = integer(default=None) # Max vertex separation

        # Sky statistics parameters:
        skystat = option('median', 'midpt', 'mean', 'mode', default='mode') # sky statistics
        dqbits = string(default='~DO_NOT_USE+NON_SCIENCE') # "good" DQ bits
        lower = float(default=None) # Lower limit of "good" pixel values
        upper = float(default=None) # Upper limit of "good" pixel values
        nclip = integer(min=0, default=5) # number of sky clipping iterations
        lsigma = float(min=0.0, default=4.0) # Lower clipping limit, in sigma
        usigma = float(min=0.0, default=4.0) # Upper clipping limit, in sigma
        binwidth = float(min=0.0, default=0.1) # Bin width for 'mode' and 'midpt' `skystat`, in sigma
    """  # noqa: E501

    reference_file_types = []

    def process(self, input):
        self.log.setLevel(logging.DEBUG)

        if isinstance(input, ModelLibrary):
            library = input
        else:
            library = ModelLibrary(input)

        self._dqbits = interpret_bit_flags(self.dqbits, flag_name_map=pixel)

        # set sky statistics:
        self._skystat = SkyStats(
            skystat=self.skystat,
            lower=self.lower,
            upper=self.upper,
            nclip=self.nclip,
            lsig=self.lsigma,
            usig=self.usigma,
            binwidth=self.binwidth,
        )

        # create a list of "Sky" Images and/or Groups:
        images = []
        with library:
            for index, model in enumerate(library):
                images.append(self._imodel2skyim(model))

            # match/compute sky values:
            match(
                images,
                skymethod=self.skymethod,
                match_down=self.match_down,
                subtract=self.subtract,
            )

            # set sky background value in each image's meta:
            for im in images:
                if isinstance(im, SkyImage):
                    self._set_sky_background(
                        im, "COMPLETE" if im.is_sky_valid else "SKIPPED"
                    )
                else:
                    for gim in im:
                        self._set_sky_background(
                            gim, "COMPLETE" if gim.is_sky_valid else "SKIPPED"
                        )
            for index, image in enumerate(images):
                library.shelve(image.meta["image_model"], index)

        return library

    def _imodel2skyim(self, image_model):
        input_image_model = image_model

        if "background" not in image_model.meta:
            image_model.meta["background"] = maker_utils.mk_sky_background(
                level=None, subtracted=None, method="None"
            )

        if self._dqbits is None:
            dqmask = np.isfinite(image_model.data).astype(dtype=np.uint8)
        else:
            dqmask = bitfield_to_boolean_mask(
                image_model.dq, self._dqbits, good_mask_value=1, dtype=np.uint8
            ) * np.isfinite(image_model.data)

        # see if 'skymatch' was previously run and raise an exception
        # if 'subtract' mode has changed compared to the previous pass:
        level = image_model.meta.background.level
        if image_model.meta.background.subtracted is None:
            if level is not None:
                # report inconsistency:
                raise ValueError(
                    "Background level was set but the "
                    "'subtracted' property is undefined (None)."
                )
            # Assume level is zero:
            level = 0.0

        else:
            if image_model.meta.background.subtracted and level is None:
                # NOTE: In principle we could assume that level is 0 and
                # possibly add a log entry documenting this, however,
                # at this moment I think it is safer to quit and...
                #
                # report inconsistency:
                raise ValueError(
                    "Background level was subtracted but the "
                    "'level' property is undefined (None)."
                )

            if image_model.meta.background.subtracted and self.subtract:
                # cannot run 'skymatch' step on already "skymatched" images
                # when 'subtract' spec is inconsistent with
                # meta.background.subtracted:
                raise ValueError(
                    "'subtract' step's specification is "
                    "inconsistent with background info already "
                    "present in image '{:s}' meta.".format(image_model.meta.filename)
                )

        wcs = deepcopy(image_model.meta.wcs)

        sky_im = SkyImage(
            image=image_model.data,
            wcs_fwd=wcs.forward_transform,
            wcs_inv=wcs.backward_transform,
            pix_area=1.0,  # TODO: pixel area
            convf=1.0,  # TODO: conv. factor to brightness
            mask=dqmask,
            id=image_model.meta.filename,  # file name?
            skystat=self._skystat,
            stepsize=self.stepsize,
            reduce_memory_usage=False,
            meta={"image_model": input_image_model},
        )

        if self.subtract:
            sky_im.sky = level

        return sky_im

    def _set_sky_background(self, sky_image, step_status):
        image = sky_image.meta["image_model"]
        sky = sky_image.sky if sky_image.sky is not None else 0

        image.meta.background.method = str(self.skymethod)
        image.meta.background.subtracted = self.subtract
        # In numpy 2, the dtypes are more carefully controlled, so to match the
        # schema the data type needs to be re-cast to float64.
        image.meta.background.level = sky

        if step_status == "COMPLETE" and self.subtract:
            image.data[...] = sky_image.image[...]

        image.meta.cal_step.skymatch = step_status



import io
import os
from pathlib import Path
from typing import Union

import roman_datamodels as rdm

from romancal.datamodels import ModelLibrary


def check(init: Union[os.PathLike, Path, io.FileIO]) -> str:
    """
    Determine the type of a file and return it as a string

    Parameters
    ----------

    init : str
        file path or file object

    Returns
    -------
    file_type: str
        a string with the file type ("asdf", "asn", "DataModel", or "ModelLibrary")

    """

    supported = ("asdf", "json", "DataModel")

    if isinstance(init, (str, os.PathLike, Path)):
        path, ext = os.path.splitext(init)
        ext = ext.strip(".")

        if not ext:
            raise ValueError(f"Input file path does not have an extension: {init}")

        if ext not in supported:  # Could be the file is zipped; try splitting again
            path, ext = os.path.splitext(path)
            ext = ext.strip(".")

            if ext not in supported:
                raise ValueError(f"Unrecognized file type for: {init}")

        if ext == "json":  # Assume json input is an association
            return "asn"

        return ext
    elif isinstance(init, rdm.DataModel):
        return "DataModel"

    elif isinstance(init, ModelLibrary):
        return "ModelLibrary"

    elif hasattr(init, "read") and hasattr(init, "seek"):
        magic = init.read(5)
        init.seek(0, 0)

        if not magic or len(magic) < 5:
            raise ValueError(f"Cannot get file type of {str(init)}")

        if magic == b"#ASDF":
            return "asdf"

        return "asn"
    else:
        raise ValueError(f"Cannot get file type of {str(init)}")



class Counter:
    """Like itertools.count but access to the current value"""

    def __init__(self, start=0, step=1, end=None):
        self.value = start
        self.step = step
        self.end = end

    def __iter__(self):
        return self

    def __next__(self):
        if self.end is not None and abs(self.value) > abs(self.end):
            raise StopIteration
        self.value += self.step
        return self.value

    def next(self):
        """python2 compatibility"""
        return self.__next__()

    def set(self, value):
        """Set new value for counter"""
        self.value = value
        return self.value



import numpy as np
from numpy.testing import assert_allclose

from romancal.refpix.data import (
    ChannelView,
    Coefficients,
    Const,
    ReferenceFFT,
    StandardView,
)

from . import reference_utils
from .conftest import RNG, Dims


def test_constants_sanity():
    """
    Sanity check on the constants used and their relation to the test data.
        SMOKE TEST for adjustment to these assumptions.
    """
    assert Const.REF == 4
    assert Const.PAD == 12
    assert Const.CHAN_WIDTH == 128
    assert Const.N_DETECT_CHAN == 32

    # Check the relationship between the constants
    assert Const.N_COLUMNS == 4096
    assert Dims.N_ROWS == 12
    assert Dims.N_COLS == 4224

    # Check that there are the correct total number of channels
    assert Dims.N_CHAN == 33 == Const.N_DETECT_CHAN + 1

    # Check the padded channel width
    assert Dims.PADDED_WIDTH == 140 == Const.CHAN_WIDTH + Const.PAD

    # Check the number of columns in the FFT
    assert Dims.N_FFT_COLS == 841


def test_data(data):
    """
    Sanity that the generated data is the correct shape and type.
    """
    assert data.shape == (Dims.N_FRAMES, Dims.N_ROWS, Dims.N_COLS)
    assert data.dtype == np.float32


class TestStandardView:
    def test_construct_from_datamodel(self, datamodel):
        """
        Test construct from the datamodel passed in
        """
        standard = StandardView.from_datamodel(datamodel)

        # StandardView's data should not be a view to the datamodel
        assert standard.data.base is None

        # Check the relationship between the standard view and the datamodel
        assert (standard.detector == datamodel.data).all()
        assert (standard.left == datamodel.border_ref_pix_left).all()
        assert (standard.right == datamodel.border_ref_pix_right).all()

        # The amp33's dtype changes because it needs to be promoted to match that
        # of the rest of the data
        assert (standard.amp33 == datamodel.amp33.astype(np.float32)).all()

    def test_update(self, datamodel):
        standard = StandardView.from_datamodel(datamodel)
        old_data = standard.data.copy()

        # Update the standard view's data
        standard.data = RNG.uniform(
            1, 100, size=(Dims.N_FRAMES, Dims.N_ROWS, Dims.N_COLS)
        ).astype(np.float32)
        assert (standard.data != old_data).any()

        old_detector = datamodel.data.copy()
        old_amp33 = datamodel.amp33.copy()
        old_left = datamodel.border_ref_pix_left.copy()
        old_right = datamodel.border_ref_pix_right.copy()
        # Update the datamodel's data
        new = standard.update(datamodel)

        # Check that new is the datamodel
        assert new is datamodel

        # Check that the datamodel's data has been updated
        assert (new.data != old_detector).any()
        assert (new.amp33 != old_amp33).any()
        assert (new.border_ref_pix_left != old_left).any()
        assert (new.border_ref_pix_right != old_right).any()

        # Check the dtype has been preserved
        assert new.data.dtype == old_detector.dtype
        assert new.amp33.dtype == old_amp33.dtype
        assert new.border_ref_pix_left.dtype == old_left.dtype
        assert new.border_ref_pix_right.dtype == old_right.dtype

        # Check the data has been updated correctly
        assert (new.data == standard.detector).all()
        assert (new.border_ref_pix_left == standard.left).all()
        assert (new.border_ref_pix_right == standard.right).all()

        # The amp33's dtype changes because it needs to be shifted to match the
        # original data's dtype
        assert (new.amp33 == standard.amp33.astype(old_amp33.dtype)).all()

    def test_create_standard_view(self, data):
        """
        Sanity that the StandardView constructs correctly
        """
        standard = StandardView(data)
        assert standard.data is data
        assert standard.offset is None

    def test_detector(self, data):
        """"""
        standard = StandardView(data)
        detector = standard.detector

        # Check that the detector is the correct shape and type
        assert detector.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.N_COLUMNS)
        assert detector.dtype == np.float32

        # Check that the detector is the correct part of the data
        assert (detector == data[:, :, : Const.N_COLUMNS]).all()

        # Check that the detector is a view of the data
        assert detector.base is data
        assert detector.base is standard.data

    def test_left(self, data):
        standard = StandardView(data)
        left = standard.left

        # Check that the left is the correct shape and type
        assert left.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.REF)
        assert left.dtype == np.float32

        # Check that the left is the correct part of the data
        assert (left == data[:, :, : Const.REF]).all()

        # Check that the left is a view of the data
        assert left.base is data
        assert left.base is standard.data

    def test_right(self, data):
        standard = StandardView(data)
        right = standard.right

        # Check that the right is the correct shape and type
        assert right.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.REF)
        assert right.dtype == np.float32

        # Check that the right is the correct part of the data
        assert (
            right == data[:, :, -Const.REF - Const.CHAN_WIDTH : -Const.CHAN_WIDTH]
        ).all()

        # Check that the right is a view of the data
        assert right.base is data
        assert right.base is standard.data

    def test_amp33(self, data):
        standard = StandardView(data)
        amp33 = standard.amp33

        # Check that the amp33 is the correct shape and type
        assert amp33.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.CHAN_WIDTH)
        assert amp33.dtype == np.float32

        # Check that the amp33 is the correct part of the data
        assert (amp33 == data[:, :, -Const.CHAN_WIDTH :]).all()

        # Check that the amp33 is a view of the data
        assert amp33.base is data
        assert amp33.base is standard.data

    def test_data_view_relations(self, standard):
        """
        Check the relationships among the data views.
        """
        # Check that the detector + amp33 is the same as the data
        assert (np.dstack([standard.detector, standard.amp33]) == standard.data).all()

        # Check that the left is the left side of the detector
        assert (standard.left == standard.detector[:, :, : Const.REF]).all()

        # Check that the right is the right side of the detector
        assert (standard.right == standard.detector[:, :, -Const.REF :]).all()

    def test_channels(self, data, offset):
        """
        Check aligning the data into channels and padding the columns.
        """
        non_view_data = data.copy()
        standard = StandardView(data, offset=offset)
        channel_view = standard.channels

        # Check the data is unchanged by the property
        assert (standard.data == non_view_data).all()

        # Check the channel_view is a ChannelView object
        assert isinstance(channel_view, ChannelView)
        assert channel_view.offset is standard.offset

        channels = channel_view.data

        # check that the data is not a view
        assert channels.base is None

        # Check that the channels are the correct shape and type
        assert channels.shape == (
            Dims.N_CHAN,
            Dims.N_FRAMES,
            Dims.N_ROWS,
            Const.CHAN_WIDTH + Const.PAD,
        )
        assert channels.dtype == np.float32

        # Check the channels
        for idx in range(0, Dims.N_CHAN):
            truth = data[:, :, idx * Const.CHAN_WIDTH : (idx + 1) * Const.CHAN_WIDTH]
            if idx % 2 == 0:
                # Even channels are not reversed
                assert (channels[idx, :, :, : Const.CHAN_WIDTH] == truth).all()
            else:
                # Odd channels are reversed
                assert (
                    channels[idx, :, :, : Const.CHAN_WIDTH] == truth[:, :, ::-1]
                ).all()

        # Check that amp33 is the last channel
        assert (standard.amp33 == channels[-1, :, :, : Const.CHAN_WIDTH]).all()

        # Check that the first part of the first channel is left
        assert (standard.left == channels[0, :, :, : Const.REF]).all()

        # Check that the last part of the last detector channel is right
        #     This is an odd indexed channel, so it is reversed, this also means
        #     the first 4 columns in the channel form are the ones corresponding
        #     to the right side of the detector (not the last 4 columns) as one
        #     might expect.
        assert (standard.right[:, :, ::-1] == channels[-2, :, :, : Const.REF]).all()

    def test_channels_regression(self, data):
        standard = StandardView(data)
        regression = data.copy()

        # Run the regression utility
        regression = reference_utils.aligned_channels(regression)

        # Check that the regression matches the channels property
        assert (standard.channels.data == regression).all()

    def test_remove_offset(self, data):
        non_view_data = data.copy()
        standard = StandardView(data)
        assert standard.offset is None

        new = standard.remove_offset()

        # Check that new object returned is the same as the original
        assert new is standard

        # Check that the data is still a view of the original
        assert new.data is data

        # Check that the offset is now set
        assert new.offset is not None
        assert new.offset.shape == (Dims.N_ROWS, Dims.N_COLS)
        assert (new.offset != 0).all()

        # Check that the data has been updated and that the offset has been removed
        assert (new.data != non_view_data).any()
        assert (new.data == (non_view_data - new.offset)).all()

        # add offset back (error accumulates due to floating point arithmetic)
        reset = StandardView(new.data + new.offset)
        assert_allclose(reset.data, non_view_data, atol=1e-5)

    def test_remove_offset_regression(self, data):
        standard = StandardView(data)

        # Copy the data, because the regression utility modifies the data in-place
        regression = data.copy()
        _, b = reference_utils.remove_linear_trends(regression, True)

        # Run the internal utility
        new = standard.remove_offset()

        # Check that the regression matches the new object
        assert (new.offset == b).all()
        assert (new.data == regression).all()

    def test_apply_offset(self, data, offset):
        standard = StandardView(data.copy(), offset)

        # Apply the offset
        new = standard.apply_offset()

        # Check the new object is the same as the original
        assert new is standard

        # Check apply offset has modified new's state
        assert new.offset is None
        assert (new.data != data).any()

        # Check that the data has been correctly modified
        for new_frame, frame in zip(new.data, data):
            assert (new_frame == (frame + offset)).all()

        # Check if no offset is passed
        standard = StandardView(data.copy())
        new = standard.apply_offset()
        assert new is standard
        assert (new.data == data).all()

    def test_apply_offset_regression(self, data, offset):
        standard = StandardView(data.copy(), offset)

        # Run the reference utility
        regression = data[:, :, : Const.N_COLUMNS].copy()
        reference_utils.restore_offsets(Dims.N_FRAMES, regression, offset)

        # Run the internal utility
        new = standard.apply_offset()

        # Check regression
        assert (new.data[:, :, : Const.N_COLUMNS] == regression).all()

    def test_offset_roundtrip(self, data):
        standard = StandardView(data.copy())

        no_offset = standard.remove_offset()
        assert no_offset.offset is not None
        assert (no_offset.data != data).any()

        reset = no_offset.apply_offset()
        assert reset.offset is None

        # Check round trip has not changed the data.
        #   The differences are due to floating point arithmetic.
        assert_allclose(reset.data, data, atol=1e-5)


class TestChannelView:
    def test_detector(self, standard):
        channels = standard.channels
        detector = channels.detector

        # Check that the detector is the correct shape and type
        assert detector.shape == (
            Const.N_DETECT_CHAN,
            Dims.N_FRAMES,
            Dims.N_ROWS,
            Dims.PADDED_WIDTH,
        )
        assert detector.dtype == np.float32

        # Check that the detector is the correct part of the data
        assert (detector == channels.data[: Const.N_DETECT_CHAN, :, :, :]).all()

        # Check that the detector is a view of the data
        assert detector.base is channels.data

    def test_left(self, standard):
        channels = standard.channels
        left = channels.left

        # Check that the left is the correct shape and type
        assert left.shape == (Dims.N_FRAMES, Dims.N_ROWS, Dims.PADDED_WIDTH)
        assert left.dtype == np.float32

        # Check the padding is correct
        assert (left[:, :, Const.REF :] == 0).all()

        # Check that the left is the correct part of the data
        assert (left[:, :, : Const.REF] == channels.data[0, :, :, : Const.REF]).all()
        assert (
            left[:, :, : Const.REF] == channels.detector[0, :, :, : Const.REF]
        ).all()
        assert (left[:, :, : Const.REF] == standard.left).all()

        # Check that the left is not a view
        assert left.base is None

    def test_right(self, standard):
        channels = standard.channels
        right = channels.right

        # Check that the right is the correct shape and type
        assert right.shape == (Dims.N_FRAMES, Dims.N_ROWS, Dims.PADDED_WIDTH)
        assert right.dtype == np.float32

        # Check the padding is correct
        assert (right[:, :, Const.REF :] == 0).all()

        # Check that the right is the correct part of the data
        assert (right[:, :, : Const.REF] == channels.data[-2, :, :, : Const.REF]).all()
        assert (
            right[:, :, : Const.REF] == channels.detector[-1, :, :, : Const.REF]
        ).all()

        # Due to the channel flip the right will need to be reversed to match the
        #   standard.right
        assert (right[:, :, : Const.REF] == standard.right[:, :, ::-1]).all()

        # Check that the right is not a view
        assert right.base is None

    def test_amp33(self, standard):
        channels = standard.channels
        amp33 = channels.amp33

        # Check that the right is the correct shape and type
        assert amp33.shape == (Dims.N_FRAMES, Dims.N_ROWS, Dims.PADDED_WIDTH)
        assert amp33.dtype == np.float32

        # Check that the right is the correct part of the data
        assert (amp33 == channels.data[-1, :, :, :]).all()
        assert (amp33[:, :, : Const.CHAN_WIDTH] == standard.amp33).all()

        # Check that the amp33 is a view
        assert amp33.base is channels.data

    def test_standard(self, standard, offset):
        channels = standard.channels
        channels.offset = offset
        non_view_data = channels.data.copy()

        standard_view = channels.standard

        # Check the channels.data is unchanged by the property
        assert (channels.data == non_view_data).all()

        # Check the standard_view is a StandardView object
        assert isinstance(standard_view, StandardView)
        assert standard_view.offset is channels.offset

        data = standard_view.data

        # Check the data is not a view of the channels.data
        assert data.base is not channels.data

        # Check that the data is the correct shape and type
        assert data.shape == (Dims.N_FRAMES, Dims.N_ROWS, Dims.N_COLS)
        assert data.dtype == np.float32

        # Check that data matches the standard data (round trip)
        assert (data == standard.data).all()

    def test_remove_trends(self, standard):
        channels = standard.channels
        non_view_data = channels.data.copy()

        new = channels.remove_trends()

        # Check that new object returned is the original
        assert new is channels

        # Check the data is not a view
        assert new.data.base is None

        # Check the method has updated the data
        assert (new.data != non_view_data).any()

    def test_remove_trends_regression(self, standard):
        channels = standard.channels

        # Copy the data, because the regression utility modifies the data in-place
        regression = channels.data.copy()
        reference_utils.remove_linear_trends_per_frame(regression, False, True)

        # Run the internal utility
        new = channels.remove_trends()

        # Check that the regression matches the new object
        assert (new.data == regression).all()

    def test_cosine_interpolate(self, standard):
        channels = standard.channels
        non_view_data = channels.data.copy()

        new = channels.cosine_interpolate()

        # Check that new object returned is the original
        assert new is channels

        # Check the data is not a view
        assert new.data.base is None

        # Check the method has updated only the amp33 data
        assert (new.amp33 != non_view_data[-1, :, :, :]).any()
        assert (new.detector == non_view_data[:-1, :, :, :]).all()

    def test_cosine_interpolate_regression(self, channels):
        # Copy the data, because the regression utility modifies the data in-place
        regression = channels.data.copy()
        reference_utils.cos_interp_reference(regression, regression.shape[1])

        # Demonstrate the regression only ran on the amp33 data
        assert (regression[-1, :, :, :] != channels.data[-1, :, :, :]).any()
        assert (regression[:-1, :, :, :] == channels.data[:-1, :, :, :]).all()

        # Run the internal utility
        new = channels.cosine_interpolate()

        # Check that the regression matches the new object
        assert (new.amp33 == regression[-1, :, :, :]).all()
        assert (new.data == regression).all()

    def test_fft_interpolate(self, channels):
        non_view_data = channels.data.copy()

        new = channels.fft_interpolate()

        # Check that the new object returned is the original
        assert new is channels

        # Check the data is not a view
        assert new.data.base is None

        # Check the method has updated only the amp33 data
        assert (new.amp33 != non_view_data[-1, :, :, :]).any()
        assert (new.detector == non_view_data[:-1, :, :, :]).all()

    def test_fft_interpolate_regression(self, channels):
        """
        Run fft interpolation regression test
        """
        non_view_data = channels.data.copy()

        # Copy the data, because the regression utility modifies the data in-place
        regression = channels.data.copy()
        amp33_regression = reference_utils.fft_interp_amp33(
            regression, regression.shape[1]
        )

        # Demonstrate that the sub-array (amp33_regression) does have changes made
        assert (amp33_regression[:, :, :] != non_view_data[-1, :, :, :]).any()

        # Demonstrate that the actual regression "data" remains unchanged instead
        # the claimed in-place change. (this would fail if in-place change was made)
        assert (regression == non_view_data).all()

        # Run the internal utility
        new = channels.fft_interpolate()

        assert (new.amp33 == amp33_regression).all()

    def test_reference_fft(self, channels):
        reference = channels.reference_fft

        # Check the new object is a ReferenceFFT object
        assert isinstance(reference, ReferenceFFT)

        # Check the FFT all have the correct shape
        shape = (Dims.N_FRAMES, (Dims.N_ROWS * Dims.PADDED_WIDTH) // 2 + 1)
        assert reference.left.shape == shape
        assert reference.right.shape == shape
        assert reference.amp33.shape == shape

        # Check the FFT are all complex dtype
        assert reference.left.dtype == np.complex64
        assert reference.right.dtype == np.complex64
        assert reference.amp33.dtype == np.complex64

        # Check the FFT are not views
        assert reference.left.base is None
        assert reference.right.base is None
        assert reference.amp33.base is None

    def test_reference_fft_regression(self, channels):
        regression = channels.data.copy()
        reference = channels.reference_fft

        # Sow that the reference property does not alter the data
        assert (regression == channels.data).all()

        # run the reference code
        left, right, amp33 = reference_utils.forward_fft(Dims.N_FRAMES, regression)

        # Copy to make sure regression utility does not modify the data in-place
        assert (regression == channels.data).all()

        # Check that the regression matches the new object
        assert (reference.left == left).all()
        assert (reference.right == right).all()
        assert (reference.amp33 == amp33).all()

    def test_correction(self, channels, coeffs):
        correction = channels.correction(coeffs)

        # Check size and dtype
        assert correction.shape == (
            Dims.N_CHAN,
            Dims.N_FRAMES,
            Dims.N_ROWS,
            Dims.PADDED_WIDTH,
        )
        assert correction.dtype == np.float32

    def test_correction_regression(self, channels, coeffs):
        # Run the reference code
        regression = reference_utils.correction(
            Dims.N_FRAMES, channels.data.copy(), coeffs.alpha, coeffs.gamma, coeffs.zeta
        )

        # Check size and dtype
        assert regression.shape == (
            Dims.N_CHAN,
            Dims.N_FRAMES,
            Dims.N_ROWS,
            Dims.PADDED_WIDTH,
        )
        assert regression.dtype == np.float64

        # Run the internal utility
        correction = channels.correction(coeffs)

        assert (correction == regression.astype(np.float32)).all()

    def test_apply_correction(self, channels, coeffs):
        non_view_standard_data = channels.standard.data.copy()
        non_view_data = channels.data.copy()

        amp33 = channels.amp33.copy()

        # Compute the applied correction
        standard = channels.apply_correction(coeffs)

        # Check the correction is the same as the standard
        assert isinstance(standard, StandardView)

        # Check the channels is updated in-place
        assert (channels.data != non_view_data).any()

        # Check the standard is different from the original
        assert (standard.data != non_view_standard_data).any()

        # Check the amp33 channel has not changed
        assert (channels.amp33 == amp33).all()

    def test_apply_correction_channels_regression(self, channels, coeffs):
        # Get regression
        regression = reference_utils.apply_correction_to_channels(
            Dims.N_FRAMES, channels.data.copy(), coeffs.alpha, coeffs.gamma, coeffs.zeta
        )

        # Check size and dtype
        assert regression.shape == (
            Dims.N_CHAN,
            Dims.N_FRAMES,
            Dims.N_ROWS,
            Const.CHAN_WIDTH,
        )
        assert regression.dtype == np.float32

        # Run the internal utility (looking at the inplace changes)
        channels.apply_correction(coeffs)

        # Check regression
        assert (channels.data[:, :, :, : Const.CHAN_WIDTH] == regression).all()

    def test_apply_correction_standard_regression(self, channels, coeffs):
        regression = reference_utils.apply_correction_to_data(
            Dims.N_FRAMES, channels.data.copy(), coeffs.alpha, coeffs.gamma, coeffs.zeta
        )

        # Check size and dtype
        assert regression.shape == (Dims.N_FRAMES, Dims.N_ROWS, Const.N_COLUMNS)
        assert regression.dtype == np.float32

        # Run the internal utility
        standard = channels.apply_correction(coeffs)

        # Check regression
        assert (standard.detector == regression).all()


class TestReferenceFFT:
    def test_correction(self, channels, coeffs):
        reference = channels.reference_fft

        left = reference.left.copy()
        right = reference.right.copy()
        amp33 = reference.amp33.copy()

        correction = reference.correction(coeffs)

        # Check size and dtype
        assert correction.shape == (
            Dims.N_CHAN,
            Dims.N_FRAMES,
            Dims.N_ROWS * Dims.PADDED_WIDTH,
        )
        assert correction.dtype == np.float64  # no dtype casting yet

        # Check the last channel is 0
        assert (correction[-1, :, :] == 0).all()

        # Check the FFT have not been modified
        assert (left == reference.left).all()
        assert (right == reference.right).all()
        assert (amp33 == reference.amp33).all()

    def test_correction_regression(self, channels, coeffs):
        reference = channels.reference_fft

        # Run reference code
        regression = reference_utils.compute_correction(
            Dims.N_ROWS,
            coeffs.alpha,
            coeffs.gamma,
            coeffs.zeta,
            Dims.N_FRAMES,
            reference.left,
            reference.right,
            reference.amp33,
        )

        # Run implementation
        correction = reference.correction(coeffs)

        assert correction.dtype == regression.dtype
        assert (correction == regression).all(), correction.dtype


class TestCoefficients:
    def test_from_ref(self, ref_pix_ref):
        coeffs = Coefficients.from_ref(ref_pix_ref)

        # Check the object is a Coefficients object
        assert isinstance(coeffs, Coefficients)

        # Check the coefficients are the same
        assert (coeffs.alpha == ref_pix_ref.alpha).all()
        assert (coeffs.gamma == ref_pix_ref.gamma).all()
        assert (coeffs.zeta == ref_pix_ref.zeta).all()



"""Apply the flux scaling"""

import logging

from roman_datamodels import datamodels

from ..datamodels import ModelLibrary
from ..stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["FluxStep"]


class FluxStep(RomanStep):
    """Apply flux scaling to count-rate data

    Parameters
    -----------
    input : str, `roman_datamodels.datamodels.DataModel`, or `~romancal.datamodels.library.ModelLibrary`
        If a string is provided, it should correspond to either a single ASDF filename
        or an association filename. Alternatively, a single DataModel instance can be
        provided instead of an ASDF filename. Multiple files can be processed via
        either an association file or wrapped by a
        `~romancal.datamodels.library.ModelLibrary`.

    Returns
    -------
    output_models : `roman_datamodels.datamodels.DataModel`, or `~romancal.datamodels.library.ModelLibrary`
        The models with flux applied.


    Notes
    -----
    Currently, the correction is done in-place; the inputs are directly modified if in-memory DataModels are input.
    """  # noqa: E501

    class_alias = "flux"

    spec = """
    """  # noqa: E501

    reference_file_types = []

    def process(self, input):
        if isinstance(input, datamodels.DataModel):
            input_models = ModelLibrary([input])
            single_model = True
        elif isinstance(input, str):
            # either a single asdf filename or an association filename
            try:
                # association filename
                input_models = ModelLibrary(input)
                single_model = False
            except Exception:
                # single ASDF filename
                input_models = ModelLibrary([datamodels.open(input)])
                single_model = True
        elif isinstance(input, ModelLibrary):
            input_models = input
            single_model = False
        else:
            raise TypeError(
                "Input must be an ASN filename, a ModelLibrary, "
                "a single ASDF filename, or a single Roman DataModel."
            )

        with input_models:
            for index, model in enumerate(input_models):
                apply_flux_correction(model)
                model.meta.cal_step.flux = "COMPLETE"
                input_models.shelve(model, index)

        if single_model:
            with input_models:
                model = input_models.borrow(0)
                input_models.shelve(model, 0, modify=False)
            return model
        return input_models


def apply_flux_correction(model):
    """Apply the flux correction

    The input model is expected to be Roman ImageModel-like.
    The photometry information is taken from `meta.photometry`.

    The model is converted in-place.

    Parameters
    ----------
    model : `ImageModel`
        The model to apply the flux correction to. The model is modified in-place.

    Notes
    -----
    The modifications to the model can result in validation issues due to change of units.
    """
    # Define the various arrays to be converted.
    DATA = ("data", "err")
    VARIANCES = ("var_rnoise", "var_poisson", "var_flat")

    if model.meta.cal_step["flux"] == "COMPLETE":
        message = (
            "Input data is already in flux units of MJy/sr."
            "\nFlux correction already applied."
        )
        log.info(message)
        return

    # Apply the correction.
    # The end goal in units is to have MJy/sr. The scale is in MJy/sr also.
    # Hence the extra factor of s/DN must be applied to cancel DN/s.
    log.debug("Flux correction being applied")
    c_mj = model.meta.photometry.conversion_megajanskys
    for data in DATA:
        model[data] = model[data] * c_mj
    for variance in VARIANCES:
        model[variance] = model[variance] * c_mj**2
